__author__ = "Simon Nilsson"

import glob
import os
import pickle
from tkinter import *
from typing import Union

from simba.bounding_box_tools.agg_boundary_stats import \
    AggBoundaryStatisticsCalculator
from simba.bounding_box_tools.boundary_statistics import \
    BoundaryStatisticsCalculator
from simba.bounding_box_tools.find_boundaries import AnimalBoundaryFinder
from simba.bounding_box_tools.visualize_boundaries import BoundaryVisualizer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.checks import check_int
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import NoChoosenMeasurementError, NoFilesFoundError
from simba.utils.lookups import get_named_colors
from simba.utils.read_write import find_all_videos_in_project, get_fn_ext


class BoundaryMenus(ConfigReader, PopUpMixin):
    """
    Launch GUI interface for extrapolating bounding boxes from pose-estimation data, and calculating
    statstics on bounding boxes and pose-estimated key-point intersections.

    :parameter str config_path: str path to SimBA project config file in Configparser format

    Notes
    ----------
    `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/anchored_rois.md>`_.

    Examples
    ----------
    >>> _ = BoundaryMenus(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini')
    """

    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path)
        PopUpMixin.__init__(
            self, title="SIMBA ANCHORED ROI (BOUNDARY BOXES ANALYSIS)", size=(750, 300)
        )
        self.named_shape_colors = get_named_colors()
        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.BBOXES.value,
        )
        self.max_animal_name_char = len(
            max([x for x in list(self.animal_bp_dict.keys())])
        )
        self.find_boundaries_btn = Button(
            self.settings_frm,
            text="FIND ANIMAL BOUNDARIES",
            command=lambda: self.__launch_find_boundaries_pop_up(),
        )
        self.visualize_boundaries_btn = Button(
            self.settings_frm,
            text="VISUALIZE BOUNDARIES",
            command=lambda: self.__launch_visualize_boundaries(),
        )
        self.boundary_statistics_btn = Button(
            self.settings_frm,
            text="CALCULATE BOUNDARY STATISTICS",
            command=lambda: self.__launch_boundary_statistics(),
        )
        self.agg_boundary_statistics_btn = Button(
            self.settings_frm,
            text="CALCULATE AGGREGATE BOUNDARY STATISTICS",
            command=lambda: self.__launch_agg_boundary_statistics(),
        )
        self.settings_frm.grid(row=0, sticky=W)
        self.find_boundaries_btn.grid(row=0, column=0, sticky=NW)
        self.visualize_boundaries_btn.grid(row=1, column=0, sticky=NW)
        self.boundary_statistics_btn.grid(row=0, column=1, sticky=NW)
        self.agg_boundary_statistics_btn.grid(row=1, column=1, sticky=NW)

    def __launch_find_boundaries_pop_up(self):
        self.find_boundaries_frm = Toplevel()
        self.find_boundaries_frm.minsize(750, 300)
        self.find_boundaries_frm.wm_title("FIND ANIMAL BOUNDARIES")
        self.find_boundaries_frm.lift()
        self.select_shape_type_frm = LabelFrame(
            self.find_boundaries_frm,
            text="SELECT SHAPE TYPE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=15,
        )
        self.shape_types = [
            "ENTIRE ANIMAL",
            "SINGLE BODY-PART SQUARE",
            "SINGLE BODY-PART CIRCLE",
        ]
        self.shape_dropdown = DropDownMenu(
            self.select_shape_type_frm, "SELECT SHAPE TYPE", self.shape_types, "15"
        )
        self.shape_dropdown.setChoices(self.shape_types[0])
        self.select_btn = Button(
            self.select_shape_type_frm,
            text="SELECT",
            command=lambda: self.__populate_find_boundaries_menu(),
        )
        self.select_shape_type_frm.grid(row=0, column=0, sticky=NW)
        self.shape_dropdown.grid(row=0, column=0, sticky=NW)
        self.select_btn.grid(row=0, column=1, sticky=NW)

    def __populate_find_boundaries_menu(self):
        if hasattr(self, "boundary_settings"):
            self.boundary_settings.destroy()
        self.selected_shape_type = self.shape_dropdown.getChoices()
        self.boundary_settings = LabelFrame(
            self.find_boundaries_frm,
            text="BOUNDARY SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=15,
        )
        boundary_settings_row_cnt = 1
        if (self.selected_shape_type == "SINGLE BODY-PART SQUARE") | (
            self.selected_shape_type == "SINGLE BODY-PART CIRCLE"
        ):
            self.animals = {}
            for animal_cnt, (name, animal_data) in enumerate(
                self.animal_bp_dict.items()
            ):
                self.animals[name] = {}
                self.animals[name]["animal_name_lbl"] = Label(
                    self.boundary_settings,
                    text=name,
                    width=self.max_animal_name_char + 5,
                )
                animal_bps = [x[:-2] for x in animal_data["X_bps"]]
                self.animals[name]["body_part_dropdown"] = DropDownMenu(
                    self.boundary_settings, "BODY-PART: ", animal_bps, " 0"
                )
                self.animals[name]["body_part_dropdown"].setChoices(animal_bps[0])
                self.animals[name]["animal_name_lbl"].grid(
                    row=boundary_settings_row_cnt, column=0, sticky=NW
                )
                self.animals[name]["body_part_dropdown"].grid(
                    row=boundary_settings_row_cnt, column=1, sticky=NW
                )
                boundary_settings_row_cnt += 1
        elif self.selected_shape_type == "ENTIRE ANIMAL":
            self.force_rectangle_var = BooleanVar()
            self.force_rectangle_cb = Checkbutton(
                self.boundary_settings,
                text="FORCE RECTANGLE",
                variable=self.force_rectangle_var,
                command=None,
            )
            self.force_rectangle_cb.grid(
                row=boundary_settings_row_cnt, column=0, sticky=NW
            )
            boundary_settings_row_cnt += 1
        self.boundary_settings.grid(row=1, column=0, sticky=W, padx=5, pady=15)
        self.parallel_offset_entry = Entry_Box(
            self.boundary_settings,
            "PARALLEL OFFSET (MM):",
            labelwidth="18",
            width=10,
            validation="numeric",
        )
        self.run_find_shapes_btn = Button(
            self.boundary_settings,
            text="RUN",
            command=lambda: self.__run_find_boundaries(),
        )
        self.parallel_offset_entry.entry_set("0")
        self.parallel_offset_entry.grid(
            row=boundary_settings_row_cnt, column=0, sticky=NW
        )
        self.run_find_shapes_btn.grid(
            row=boundary_settings_row_cnt + 1, column=0, sticky=NW
        )

    def __run_find_boundaries(self):
        body_parts, force_rectangle = None, False
        if self.selected_shape_type == "ENTIRE ANIMAL":
            force_rectangle = self.force_rectangle_var.get()
            body_parts = None
        elif (self.selected_shape_type == "SINGLE BODY-PART SQUARE") | (
            self.selected_shape_type == "SINGLE BODY-PART CIRCLE"
        ):
            body_parts = {}
            for animal, animal_data in self.animals.items():
                body_parts[animal] = self.animals[animal][
                    "body_part_dropdown"
                ].getChoices()
            force_rectangle = False
        parallel_offset = self.parallel_offset_entry.entry_get
        check_int(name="PARALLEL OFFSET", value=parallel_offset)
        boundary_finder = AnimalBoundaryFinder(
            config_path=self.config_path,
            roi_type=self.selected_shape_type,
            body_parts=body_parts,
            force_rectangle=force_rectangle,
            parallel_offset=int(parallel_offset),
        )
        boundary_finder.run()

    def __launch_visualize_boundaries(self):
        self.anchored_roi_path = os.path.join(
            self.project_path, "logs", "anchored_rois.pickle"
        )
        if not os.path.isfile(self.anchored_roi_path):
            raise NoFilesFoundError(
                msg="No anchored ROI found in {}.".format(self.anchored_roi_path)
            )
        with open(self.anchored_roi_path, "rb") as fp:
            self.roi_data = pickle.load(fp)
        videos_in_project = find_all_videos_in_project(videos_dir=self.video_dir)
        videos_with_data = list(self.roi_data.keys())
        if len(videos_in_project) == 0:
            raise NoFilesFoundError(msg="Zero video files found in SimBA project")
        video_names = []
        for file_path in videos_in_project:
            _, name, _ = get_fn_ext(filepath=file_path)
            video_names.append(name)
        sets_w_data_and_video = list(set(videos_with_data).intersection(video_names))
        if len(sets_w_data_and_video) == 0:
            raise NoFilesFoundError(
                msg="SIMBA ERROR: Zero video files found with calculated anchored ROIs in SimBA project"
            )
        self.viz_boundaries_frm = Toplevel()
        self.viz_boundaries_frm.minsize(600, 150)
        self.viz_boundaries_frm.wm_title("VISUALIZE ANIMAL BOUNDARIES")
        self.video_settings_frm = LabelFrame(
            self.viz_boundaries_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=15,
        )
        self.roi_attr_frm = LabelFrame(
            self.viz_boundaries_frm,
            text="ROI ATTRIBUTES",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=15,
        )
        self.select_video_dropdown = DropDownMenu(
            self.video_settings_frm, "SELECT VIDEO: ", sets_w_data_and_video, " 0"
        )
        self.select_video_dropdown.setChoices(sets_w_data_and_video[0])
        self.run_visualize_roi = Button(
            self.viz_boundaries_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="blue",
            command=lambda: self.__run_boundary_visualization(),
        )
        self.include_keypoints_var = BooleanVar()
        self.convert_to_grayscale_var = BooleanVar()
        self.highlight_intersections_var = BooleanVar()
        self.enable_attr_var = BooleanVar()
        self.include_keypoints_cb = Checkbutton(
            self.video_settings_frm,
            text="INCLUDE KEY-POINTS",
            variable=self.include_keypoints_var,
            command=None,
        )
        self.convert_to_grayscale_cb = Checkbutton(
            self.video_settings_frm,
            text="GREYSCALE",
            variable=self.convert_to_grayscale_var,
            command=None,
        )
        self.highlight_intersections_cb = Checkbutton(
            self.video_settings_frm,
            text="HIGHLIGHT INTERSECTIONS",
            variable=self.highlight_intersections_var,
            command=None,
        )
        self.enable_roi_attr_cb = Checkbutton(
            self.video_settings_frm,
            text="ENABLE USER-DEFINED ROI ATTRIBUTES",
            variable=self.enable_attr_var,
            command=self.__enable_roi_attributes,
        )
        self.video_settings_frm.grid(row=0, sticky=NW)
        self.select_video_dropdown.grid(row=0, column=0, sticky=NW)
        self.include_keypoints_cb.grid(row=1, column=0, sticky=NW)
        self.convert_to_grayscale_cb.grid(row=2, column=0, sticky=NW)
        self.highlight_intersections_cb.grid(row=3, column=0, sticky=NW)
        self.enable_roi_attr_cb.grid(row=4, column=0, sticky=NW)
        self.run_visualize_roi.grid(row=5, column=0, sticky=NW)
        self.animal_attr_dict = {}

        Label(
            self.roi_attr_frm,
            text="ANIMAL",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            width=self.max_animal_name_char + 10,
        ).grid(row=0, column=0, sticky=N)
        Label(
            self.roi_attr_frm,
            text="ROI COLOR",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            width=self.max_animal_name_char + 10,
        ).grid(row=0, column=1, sticky=N)
        Label(
            self.roi_attr_frm,
            text="ROI THICKNESS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            width=self.max_animal_name_char + 10,
        ).grid(row=0, column=2, sticky=N)
        Label(
            self.roi_attr_frm,
            text="KEY-POINT SIZE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            width=self.max_animal_name_char + 10,
        ).grid(row=0, column=3, sticky=N)
        Label(
            self.roi_attr_frm,
            text="HIGHLIGHT COLOR",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            width=self.max_animal_name_char + 10,
        ).grid(row=0, column=4, sticky=N)
        Label(
            self.roi_attr_frm,
            text="HIGHLIGHT THICKNESS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            width=self.max_animal_name_char + 10,
        ).grid(row=0, column=5, sticky=N)
        for cnt, animal_name in enumerate(self.animal_bp_dict.keys()):
            self.animal_attr_dict[animal_name] = {}
            self.animal_attr_dict[animal_name]["label"] = Label(
                self.roi_attr_frm, text=animal_name, width=self.max_animal_name_char
            )
            self.animal_attr_dict[animal_name]["clr_dropdown"] = DropDownMenu(
                self.roi_attr_frm, "", self.named_shape_colors, "10", command=None
            )
            self.animal_attr_dict[animal_name]["clr_dropdown"].setChoices(
                self.named_shape_colors[cnt]
            )
            self.animal_attr_dict[animal_name]["thickness_dropdown"] = DropDownMenu(
                self.roi_attr_frm, "", list(range(1, 10)), "10", command=None
            )
            self.animal_attr_dict[animal_name]["thickness_dropdown"].setChoices(1)
            self.animal_attr_dict[animal_name]["circle_size_dropdown"] = DropDownMenu(
                self.roi_attr_frm, "", list(range(1, 10)), "10", command=None
            )
            self.animal_attr_dict[animal_name]["circle_size_dropdown"].setChoices(1)
            self.animal_attr_dict[animal_name]["highlight_clr_dropdown"] = DropDownMenu(
                self.roi_attr_frm, "", self.named_shape_colors, "10", command=None
            )
            self.animal_attr_dict[animal_name]["highlight_clr_dropdown"].setChoices(
                "Red"
            )
            self.animal_attr_dict[animal_name]["highlight_clr_thickness"] = (
                DropDownMenu(
                    self.roi_attr_frm, "", list(range(1, 10)), "10", command=None
                )
            )
            self.animal_attr_dict[animal_name]["highlight_clr_thickness"].setChoices(5)
            self.animal_attr_dict[animal_name]["label"].grid(
                row=cnt + 1, column=0, sticky=NW
            )
            self.animal_attr_dict[animal_name]["clr_dropdown"].grid(
                row=cnt + 1, column=1, sticky=NW
            )
            self.animal_attr_dict[animal_name]["thickness_dropdown"].grid(
                row=cnt + 1, column=2, sticky=NW
            )
            self.animal_attr_dict[animal_name]["circle_size_dropdown"].grid(
                row=cnt + 1, column=3, sticky=NW
            )
            self.animal_attr_dict[animal_name]["highlight_clr_dropdown"].grid(
                row=cnt + 1, column=4, sticky=NW
            )
            self.animal_attr_dict[animal_name]["highlight_clr_thickness"].grid(
                row=cnt + 1, column=5, sticky=NW
            )
        self.roi_attr_frm.grid(row=3, column=0, sticky=NW)
        self.__enable_roi_attributes()

    def __enable_roi_attributes(self):
        if self.enable_attr_var.get():
            for animal_name in self.animal_attr_dict.keys():
                self.animal_attr_dict[animal_name]["clr_dropdown"].enable()
                self.animal_attr_dict[animal_name]["thickness_dropdown"].enable()
                self.animal_attr_dict[animal_name]["circle_size_dropdown"].enable()
                self.animal_attr_dict[animal_name]["highlight_clr_dropdown"].enable()
                self.animal_attr_dict[animal_name]["highlight_clr_thickness"].enable()
        else:
            for animal_name in self.animal_attr_dict.keys():
                self.animal_attr_dict[animal_name]["clr_dropdown"].disable()
                self.animal_attr_dict[animal_name]["thickness_dropdown"].disable()
                self.animal_attr_dict[animal_name]["circle_size_dropdown"].disable()
                self.animal_attr_dict[animal_name]["highlight_clr_dropdown"].disable()
                self.animal_attr_dict[animal_name]["highlight_clr_thickness"].disable()

    def __run_boundary_visualization(self):
        include_keypoints = self.include_keypoints_var.get()
        greyscale = self.convert_to_grayscale_var.get()
        highlight_intersections = self.highlight_intersections_var.get()
        roi_attr = {}
        if self.enable_attr_var.get():
            for cnt, animal_name in enumerate(self.animal_bp_dict.keys()):
                roi_attr[animal_name] = {}
                roi_attr[animal_name]["bbox_clr"] = self.named_shape_colors[
                    self.animal_attr_dict[animal_name]["clr_dropdown"].getChoices()
                ]
                roi_attr[animal_name]["bbox_thickness"] = int(
                    self.animal_attr_dict[animal_name][
                        "thickness_dropdown"
                    ].getChoices()
                )
                roi_attr[animal_name]["keypoint_size"] = int(
                    self.animal_attr_dict[animal_name][
                        "circle_size_dropdown"
                    ].getChoices()
                )
                roi_attr[animal_name]["highlight_clr"] = self.named_shape_colors[
                    self.animal_attr_dict[animal_name][
                        "highlight_clr_dropdown"
                    ].getChoices()
                ]
                roi_attr[animal_name]["highlight_clr_thickness"] = int(
                    self.animal_attr_dict[animal_name][
                        "highlight_clr_thickness"
                    ].getChoices()
                )
        else:
            roi_attr = None

        video_visualizer = BoundaryVisualizer(
            config_path=self.config_path,
            video_name=self.select_video_dropdown.getChoices(),
            include_key_points=include_keypoints,
            greyscale=greyscale,
            show_intersections=highlight_intersections,
            roi_attributes=roi_attr,
        )

        video_visualizer.run()

    def __launch_boundary_statistics(self):
        self.anchored_roi_path = os.path.join(
            self.project_path, "logs", "anchored_rois.pickle"
        )
        if not os.path.isfile(self.anchored_roi_path):
            raise NoFilesFoundError(
                msg="SIMBA ERROR: No anchored ROI found in {}.".format(
                    self.anchored_roi_path
                )
            )
        self.statistics_frm = Toplevel()
        self.statistics_frm.minsize(400, 150)
        self.statistics_frm.wm_title("ANIMAL ANCHORED ROI STATISTICS")
        self.settings_frm = LabelFrame(
            self.statistics_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=15,
        )
        self.file_type_frm = LabelFrame(
            self.statistics_frm,
            text="OUTPUT FILE TYPE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=15,
        )
        self.roi_intersections_var = BooleanVar(value=True)
        self.roi_keypoint_intersections_var = BooleanVar(value=True)
        self.roi_intersections_cb = Checkbutton(
            self.settings_frm,
            text="ROI-ROI INTERSECTIONS",
            variable=self.roi_intersections_var,
            command=None,
        )
        self.roi_keypoint_intersections_cb = Checkbutton(
            self.settings_frm,
            text="ROI-KEYPOINT INTERSECTIONS",
            variable=self.roi_keypoint_intersections_var,
            command=None,
        )
        self.out_file_type = StringVar(value="CSV")
        input_csv_rb = Radiobutton(
            self.file_type_frm, text=".csv", variable=self.out_file_type, value="CSV"
        )
        input_parquet_rb = Radiobutton(
            self.file_type_frm,
            text=".parquet",
            variable=self.out_file_type,
            value="PARQUET",
        )
        input_pickle_rb = Radiobutton(
            self.file_type_frm,
            text=".pickle",
            variable=self.out_file_type,
            value="PICKLE",
        )
        self.run_statistics = Button(
            self.statistics_frm, text="RUN", command=lambda: self.__run_statistics()
        )
        self.settings_frm.grid(row=0, sticky=NW)
        self.file_type_frm.grid(row=1, sticky=NW)
        self.roi_intersections_cb.grid(row=0, column=0, sticky=NW)
        self.roi_keypoint_intersections_cb.grid(row=1, column=0, sticky=NW)
        input_csv_rb.grid(row=0, column=0, sticky=NW)
        input_parquet_rb.grid(row=1, column=0, sticky=NW)
        input_pickle_rb.grid(row=2, column=0, sticky=NW)
        self.run_statistics.grid(row=3, column=0, sticky=NW)

    def __run_statistics(self):
        save_format = self.out_file_type.get()
        roi_intersections = self.roi_intersections_var.get()
        roi_keypoint_intersections = self.roi_keypoint_intersections_var.get()
        if (not roi_intersections) and (not roi_keypoint_intersections):
            raise NoChoosenMeasurementError()
        statistics_calculator = BoundaryStatisticsCalculator(
            config_path=self.config_path,
            roi_intersections=roi_intersections,
            roi_keypoint_intersections=roi_keypoint_intersections,
            save_format=save_format,
        )
        statistics_calculator.save_results()

    def __launch_agg_boundary_statistics(self):
        self.data_path = os.path.join(self.project_path, "csv", "anchored_roi_data")
        if not os.path.isdir(self.data_path):
            raise NoFilesFoundError(
                msg="SIMBA ERROR: No anchored ROI statistics found in {}.".format(
                    self.anchored_roi_path
                )
            )
        self.main_agg_statistics_frm = Toplevel()
        self.main_agg_statistics_frm.minsize(400, 175)
        self.main_agg_statistics_frm.wm_title(
            "ANIMAL ANCHORED ROI AGGREGATE STATISTICS"
        )
        self.agg_settings_frm = LabelFrame(
            self.main_agg_statistics_frm,
            text="SETTINGS",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )

        self.interaction_time = BooleanVar(value=True)
        self.interaction_bout_cnt = BooleanVar(value=True)
        self.interaction_bout_mean = BooleanVar(value=True)
        self.interaction_bout_median = BooleanVar(value=True)
        self.detailed_interaction_data_var = BooleanVar(value=True)
        self.interaction_time_cb = Checkbutton(
            self.agg_settings_frm,
            text="INTERACTION TIME (s)",
            variable=self.interaction_time,
            command=None,
        )
        self.interaction_bout_cnt_cb = Checkbutton(
            self.agg_settings_frm,
            text="INTERACTION BOUT COUNT",
            variable=self.interaction_bout_cnt,
            command=None,
        )
        self.interaction_bout_mean_cb = Checkbutton(
            self.agg_settings_frm,
            text="INTERACTION BOUT TIME MEAN (s)",
            variable=self.interaction_bout_mean,
            command=None,
        )
        self.interaction_bout_median_cb = Checkbutton(
            self.agg_settings_frm,
            text="INTERACTION BOUT TIME MEDIAN (s)",
            variable=self.interaction_bout_median,
            command=None,
        )
        self.detailed_interaction_data_cb = Checkbutton(
            self.agg_settings_frm,
            text="DETAILED INTERACTIONS TABLE",
            variable=self.detailed_interaction_data_var,
            command=None,
        )
        self.minimum_bout_entry_box = Entry_Box(
            self.agg_settings_frm,
            "MINIMUM BOUT LENGTH (MS):",
            labelwidth="25",
            width=10,
            validation="numeric",
        )
        self.run_btn = Button(
            self.main_agg_statistics_frm,
            text="CALCULATE AGGREGATE STATISTICS",
            command=lambda: self._run_agg_stats(),
        )

        self.agg_settings_frm.grid(row=0, sticky=NW)
        self.interaction_time_cb.grid(row=0, column=0, sticky=NW)
        self.interaction_bout_cnt_cb.grid(row=1, column=0, sticky=NW)
        self.interaction_bout_mean_cb.grid(row=2, column=0, sticky=NW)
        self.interaction_bout_median_cb.grid(row=3, column=0, sticky=NW)
        self.detailed_interaction_data_cb.grid(row=4, column=0, sticky=NW)
        self.minimum_bout_entry_box.grid(row=5, column=0, sticky=NW)
        self.run_btn.grid(row=1, column=0, sticky=NW)

    def _run_agg_stats(self):
        measures = []
        for cb, name in zip(
            [
                self.interaction_time,
                self.interaction_bout_cnt,
                self.interaction_bout_mean,
                self.interaction_bout_median,
                self.detailed_interaction_data_var,
            ],
            [
                "INTERACTION TIME (s)",
                "INTERACTION BOUT COUNT",
                "INTERACTION BOUT TIME MEAN (s)",
                "INTERACTION BOUT TIME MEDIAN (s)",
                "DETAILED INTERACTIONS TABLE",
            ],
        ):
            if cb.get():
                measures.append(name)
        min_bout = self.minimum_bout_entry_box.entry_get
        check_int(name="MIN BOUT LENGTH", value=min_bout)
        if min_bout == "":
            min_bout = 0
        if len(measures) == 0:
            raise NoChoosenMeasurementError()

        agg_stats_calculator = AggBoundaryStatisticsCalculator(
            config_path=self.config_path,
            measures=measures,
            shortest_allowed_interaction=int(min_bout),
        )
        agg_stats_calculator.run()
        agg_stats_calculator.save()


# test = BoundaryMenus(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.main_frm.mainloop()



from tkinter import *

from simba.data_processors.agg_clf_calculator import AggregateClfCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.errors import (NoChoosenClassifierError,
                                NoChoosenMeasurementError)


class ClfDescriptiveStatsPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(
            self,
            title="ANALYZE CLASSIFICATIONS: DESCRIPTIVE STATISTICS",
            size=(400, 500),
        )
        ConfigReader.__init__(self, config_path=config_path)
        measures_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="MEASUREMENTS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ANALYZE_ML_RESULTS.value,
        )
        clf_frm = LabelFrame(
            self.main_frm,
            text="CLASSIFIERS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        detailed_data_frm = LabelFrame(
            self.main_frm,
            text="DETAILED BOUT DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        video_meta_data_frm = LabelFrame(
            self.main_frm,
            text="METADATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        output_options_frm = LabelFrame(
            self.main_frm,
            text="OUTPUT OPTIONS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.measurements_var_dict, self.clf_var_dict = {}, {}
        cbox_titles = Options.CLF_DESCRIPTIVES_OPTIONS.value
        for cnt, title in enumerate(cbox_titles):
            self.measurements_var_dict[title] = BooleanVar()
            cbox = Checkbutton(
                measures_frm, text=title, variable=self.measurements_var_dict[title]
            )
            cbox.grid(row=cnt, sticky=NW)
        for cnt, clf_name in enumerate(self.clf_names):
            self.clf_var_dict[clf_name] = BooleanVar()
            cbox = Checkbutton(
                clf_frm, text=clf_name, variable=self.clf_var_dict[clf_name]
            )
            cbox.grid(row=cnt, sticky=NW)
        run_button = Button(
            self.main_frm, text="Run", command=lambda: self.run_descriptive_analysis()
        )
        measures_frm.grid(row=0, sticky=NW)
        clf_frm.grid(row=1, sticky=NW)
        (
            self.metadata_frm_cnt_var,
            self.metadata_video_length_var,
            self.detailed_bout_var,
            self.transpose_output_var,
        ) = (BooleanVar(), BooleanVar(), BooleanVar(), BooleanVar())
        metadata_frm_cnt_cb = Checkbutton(
            video_meta_data_frm, text="Frame count", variable=self.metadata_frm_cnt_var
        )
        metadata_frm_video_length_cb = Checkbutton(
            video_meta_data_frm,
            text="Video length (s)",
            variable=self.metadata_video_length_var,
        )
        detailed_bout_cb = Checkbutton(
            detailed_data_frm,
            text="Detailed bout data",
            variable=self.detailed_bout_var,
        )
        transpose_output_cb = Checkbutton(
            output_options_frm,
            text="Transpose output",
            variable=self.transpose_output_var,
        )
        detailed_data_frm.grid(row=2, sticky=NW)
        detailed_bout_cb.grid(row=0, column=0, sticky=NW)
        transpose_output_cb.grid(row=0, column=0, sticky=NW)
        video_meta_data_frm.grid(row=3, sticky=NW)
        output_options_frm.grid(row=4, sticky=NW)
        metadata_frm_cnt_cb.grid(row=0, column=0, sticky=NW)
        metadata_frm_video_length_cb.grid(row=1, column=0, sticky=NW)
        run_button.grid(row=5, sticky=NW)
        self.main_frm.mainloop()

    def run_descriptive_analysis(self):
        measurement_lst, clf_list, meta_data_lst = [], [], []
        for name, val in self.measurements_var_dict.items():
            if val.get():
                measurement_lst.append(name)
        for name, val in self.clf_var_dict.items():
            if val.get():
                clf_list.append(name)
        if len(measurement_lst) == 0:
            raise NoChoosenMeasurementError(source=self.__class__.__name__)
        if len(clf_list) == 0:
            raise NoChoosenClassifierError(source=self.__class__.__name__)
        if self.metadata_frm_cnt_var.get():
            meta_data_lst.append("Frame count")
        if self.metadata_video_length_var.get():
            meta_data_lst.append("Video length (s)")
        data_log_analyzer = AggregateClfCalculator(
            config_path=self.config_path,
            data_measures=measurement_lst,
            video_meta_data=meta_data_lst,
            classifiers=clf_list,
            detailed_bout_data=self.detailed_bout_var.get(),
            transpose=self.transpose_output_var.get(),
        )
        data_log_analyzer.run()
        data_log_analyzer.save()


# ClfDescriptiveStatsPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# ClfDescriptiveStatsPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/raph/project_folder/project_config.ini')



import pytest
import os, shutil, glob
from simba.utils.read_write import get_fn_ext
from simba.utils.checks import check_file_exist_and_readable
from simba.video_processors.batch_process_create_ffmpeg_commands import FFMPEGCommandCreator
from simba.video_processors.video_processing import (change_img_format,
                                                     extract_frame_range,
                                                     change_single_video_fps,
                                                     change_fps_of_multiple_videos,
                                                     convert_video_powerpoint_compatible_format,
                                                     convert_to_mp4,
                                                     video_to_greyscale,
                                                     superimpose_frame_count,
                                                     clip_video_in_range,
                                                     downsample_video,
                                                     gif_creator,
                                                     batch_convert_video_format,
                                                     batch_create_frames,
                                                     extract_frames_single_video,
                                                     multi_split_video,
                                                     video_concatenator)

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

@pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/videos/Together_1.avi'])
def config_video_path_args(request):
    return request


@pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/videos'])
def config_video_dir_args(request):
    return request

def copy_file(file_path):
    test_file_copy_path = os.path.join(os.path.dirname(file_path), 'test.mp4')
    if os.path.isfile(test_file_copy_path): os.remove(test_file_copy_path)
    shutil.copy(file_path, test_file_copy_path)
    return test_file_copy_path

@pytest.mark.parametrize('json_path', ['tests/data/test_projects/two_c57/video_processing/batch_process_log.json'])
def test_ffmpeg_command_creator(json_path):
    ffmpeg_executor = FFMPEGCommandCreator(json_path=json_path)
    ffmpeg_executor.crop_videos()
    ffmpeg_executor.clip_videos()
    ffmpeg_executor.downsample_videos()
    ffmpeg_executor.apply_fps()
    ffmpeg_executor.apply_grayscale()
    ffmpeg_executor.apply_frame_count()
    ffmpeg_executor.apply_clahe()
    ffmpeg_executor.move_all_processed_files_to_output_folder()
    
 
@pytest.mark.parametrize('directory, file_type_in, file_type_out', [['tests/data/test_projects/two_c57/video_processing/test_imgs', 'png', 'bmp']])
def test_change_img_format(directory, file_type_in, file_type_out):
    test_dir_copy_path = os.path.join(os.path.dirname(directory), 'test_out')
    if os.path.isdir(test_dir_copy_path): shutil.rmtree(test_dir_copy_path)
    shutil.copytree(directory, test_dir_copy_path)
    change_img_format(directory=test_dir_copy_path, file_type_in=file_type_in, file_type_out=file_type_out)


@pytest.mark.parametrize('start_frame, end_frame', [[1, 5]])
def test_extract_frame_range(config_video_path_args, start_frame, end_frame):
    extract_frame_range(file_path=config_video_path_args.param, start_frame=start_frame, end_frame=end_frame)
    directory, video_name, _ = get_fn_ext(config_video_path_args.param)
    saved_files = glob.glob(os.path.join(directory, f'{video_name}_frames') + '/*.png')
    assert len(saved_files) == (end_frame - start_frame) + 1
    for i in saved_files: check_file_exist_and_readable(i)

@pytest.mark.parametrize('time',  [1])
def test_remove_beginning_of_video(config_video_path_args, time):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    #remove_beginning_of_video(file_path=copied_file_path, time=time)

@pytest.mark.parametrize('fps', [10])
def test_change_single_video_fps(config_video_path_args, fps):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    change_single_video_fps(file_path=copied_file_path, fps=fps)

@pytest.mark.parametrize('directory, fps', [['tests/data/test_projects/two_c57/project_folder/videos/', 10]])
def test_change_fps_of_multiple_videos(directory, fps):
    test_dir_copy_path = os.path.join(os.path.dirname(directory), 'test_out')
    if os.path.isdir(test_dir_copy_path): shutil.rmtree(test_dir_copy_path)
    shutil.copytree(directory, test_dir_copy_path)
    change_fps_of_multiple_videos(directory=test_dir_copy_path, fps=fps)

def test_convert_video_powerpoint_compatible_format(config_video_path_args):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    convert_video_powerpoint_compatible_format(file_path=copied_file_path)

def test_convert_to_mp4(config_video_path_args):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    convert_to_mp4(file_path=copied_file_path)

def test_video_to_greyscale(config_video_path_args):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    video_to_greyscale(file_path=copied_file_path)

def test_superimpose_frame_count(config_video_path_args):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    superimpose_frame_count(file_path=copied_file_path)

@pytest.mark.parametrize('start_time, end_time', [['00:00:01', '00:00:05']])
def test_clip_video_in_range(config_video_path_args, start_time, end_time):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    clip_video_in_range(file_path=copied_file_path, start_time=start_time, end_time=end_time)

@pytest.mark.parametrize('width, height', [[600, 300]])
def test_downsample_video(config_video_path_args, width, height):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    downsample_video(file_path=copied_file_path, video_height=height, video_width=width)

@pytest.mark.parametrize('start_time, duration, width', [[1, 3, 600]])
def test_gif_creator(config_video_path_args, start_time, duration, width):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    gif_creator(file_path=copied_file_path, width=width, start_time=start_time, duration=duration)

@pytest.mark.parametrize('start_time, duration, width', [[1, 3, 600]])
def test_gif_creator(config_video_path_args, start_time, duration, width):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    gif_creator(file_path=copied_file_path, width=width, start_time=start_time, duration=duration)

@pytest.mark.parametrize('input_format, output_format', [['avi', 'mp4']])
def test_batch_convert_video_format(config_video_dir_args, input_format, output_format):
    test_dir_copy_path = os.path.join(os.path.dirname(config_video_dir_args.param), 'test_out')
    if os.path.isdir(test_dir_copy_path): shutil.rmtree(test_dir_copy_path)
    shutil.copytree(config_video_dir_args.param, test_dir_copy_path)
    batch_convert_video_format(directory=test_dir_copy_path, input_format=input_format, output_format=output_format)

@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="LONG RUNNING TIME.")
def test_batch_create_frames(config_video_dir_args):
    test_dir_copy_path = os.path.join(os.path.dirname(config_video_dir_args.param), 'test_out')
    if os.path.isdir(test_dir_copy_path): shutil.rmtree(test_dir_copy_path)
    shutil.copytree(config_video_dir_args.param, test_dir_copy_path)
    batch_create_frames(directory=test_dir_copy_path)

@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="LONG RUNNING TIME.")
def test_extract_frames_single_video(config_video_path_args):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    extract_frames_single_video(file_path=copied_file_path)

@pytest.mark.parametrize('start_times, end_times', [(['00:00:01', '00:00:03'], ['00:00:04', '00:00:05'])])
def test_multi_split_video(config_video_path_args, start_times, end_times):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    multi_split_video(file_path=copied_file_path, start_times=start_times, end_times=end_times)

@pytest.mark.parametrize('horizontal', [True, False])
def test_video_concatenator(config_video_path_args, horizontal):
    copied_file_path = copy_file(file_path=config_video_path_args.param)
    _, video_name, _ = get_fn_ext(filepath=copied_file_path)
    video_concatenator(video_one_path=copied_file_path, video_two_path=copied_file_path, resolution=video_name, horizontal=horizontal)
    





__author__ = "Simon Nilsson"
import os

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.utils.errors import NoROIDataError


class AppendROIFeaturesByBodyPartPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise NoROIDataError(
                msg="SIMBA ERROR: No ROIs have been defined. Please define ROIs before appending ROI-based features",
                source=self.__class__.__name__,
            )
        PopUpMixin.__init__(
            self, config_path=config_path, title="APPEND ROI FEATURES: BY BODY-PARTS"
        )
        self.create_choose_number_of_body_parts_frm(
            project_body_parts=self.project_bps, run_function=self.run
        )
        self.main_frm.mainloop()

    def run(self):
        settings = {}
        settings["body_parts"] = {}
        settings["threshold"] = 0.00
        for bp_cnt, bp_dropdown in self.body_parts_dropdowns.items():
            settings["body_parts"][
                f"animal_{str(bp_cnt+1)}_bp"
            ] = bp_dropdown.getChoices()
        roi_feature_creator = ROIFeatureCreator(
            config_path=self.config_path, settings=settings
        )
        roi_feature_creator.run()
        roi_feature_creator.save()


# _ = AppendROIFeaturesByBodyPartPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')


# _ = AppendROIFeaturesByBodyPartPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



import os
import random
from datetime import datetime

import pandas as pd
# from simba.unsupervised.misc import (check_that_directory_is_empty,
#                                      check_directory_exists,
#                                      read_pickle,
#                                      write_pickle,
#                                      define_scaler,
#                                      drop_low_variance_fields,
#                                      find_low_variance_fields,
#                                      scaler_transform)
from sklearn.manifold import TSNE

import simba
from simba.misc_tools import SimbaTimer, check_file_exist_and_readable
from simba.utils.enums import Paths


class TSNEGridSearch(object):
    def __init__(self, data_path: str, save_dir: str):
        pass

    #     self.datetime = datetime.now().strftime('%Y%m%d%H%M%S')
    #     self.data_path = data_path
    #     self.save_dir = save_dir
    #     check_file_exist_and_readable(file_path=self.data_path)
    #     check_directory_exists(directory=self.save_dir)
    #     check_that_directory_is_empty(directory=self.save_dir)
    #     model_names_dir = os.path.join(os.path.dirname(simba.__file__), Paths.UNSUPERVISED_MODEL_NAMES.value)
    #     self.model_names = list(pd.read_parquet(model_names_dir)['NAMES'])
    #     self.timer = SimbaTimer()
    #     self.timer.start_timer()
    #
    # def fit(self,
    #         hyperparameters: dict):
    #     self.hyp, self.low_var_cols = hyperparameters, []
    #     self.data = read_pickle(data_path=self.data_path)
    #     self.original_feature_names = self.data['DATA'].columns
    #     if self.hyp['variance']:
    #         self.low_var_cols = find_low_variance_fields(data=self.data['DATA'], variance=self.hyp['variance'])
    #         self.data['DATA'] = drop_low_variance_fields(data=self.data['DATA'], fields=self.low_var_cols)
    #     self.out_feature_names = [x for x in self.original_feature_names if x not in self.low_var_cols]
    #     self.scaler = define_scaler(scaler_name=self.hyp['scaler'])
    #     self.scaler.fit(self.data['DATA'])
    #     self.scaled_data = scaler_transform(data=self.data['DATA'],scaler=self.scaler)
    #     self.__fit_tsne()
    #
    # def __fit_tsne(self):
    #     self.model_timer = SimbaTimer()
    #     self.model_timer.start_timer()
    #     self.results = {}
    #     self.results['SCALER'] = self.scaler
    #     self.results['LOW_VARIANCE_FIELDS'] = self.low_var_cols
    #     self.results['ORIGINAL_FEATURE_NAMES'] = self.original_feature_names
    #     self.results['OUT_FEATURE_NAMES'] = self.out_feature_names
    #     self.results['VIDEO NAMES'] = self.data['VIDEO_NAMES']
    #     self.results['START FRAME'] = self.data['START_FRAME']
    #     self.results['END FRAME'] = self.data['END_FRAME']
    #     self.results['POSE'] = self.data['POSE']
    #     self.results['DATA'] = self.scaler
    #     self.results['CLASSIFIER'] = self.data['CLF']
    #     self.results['CLASSIFIER PROBABILITY'] = self.data['CLF_PROBABILITY']
    #
    #     for h_cnt, perplexity in enumerate(self.hyp['perplexity']):
    #         self.h_cnt = h_cnt
    #         self.parameters = {'perplexity': perplexity}
    #         embedder = TSNE(n_components=2, perplexity=perplexity)
    #         embedder.fit(self.scaled_data.values)
    #         self.results['MODEL'] = embedder
    #         self.results['HASH'] = random.sample(self.model_names, 1)[0]
    #         self.results['TYPE'] = 'TSNE'
    #         write_pickle(data=self.results, save_path=os.path.join(self.save_dir, '{}.pickle'.format(self.results['HASH'])))
    #     self.timer.stop_timer()
    #     print('SIMBA COMPLETE: {} TSNE model(s) saved in {} (elapsed time: {}s)'.format(str(len(self.hyp['perplexity'])), self.save_dir, self.timer.elapsed_time_str))



__author__ = "Simon Nilsson"

""" Tkinter pop-up classes for unsupervised ML"""

import glob
from tkinter import *

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.ui.tkinter_functions import (DropDownMenu, Entry_Box, FileSelect,
                                        FolderSelect)
from simba.unsupervised.cluster_statistics import (
    ClusterFrequentistCalculator, ClusterXAICalculator,
    EmbeddingCorrelationCalculator)
from simba.unsupervised.cluster_visualizer import ClusterVisualizer
from simba.unsupervised.data_extractor import DataExtractor
from simba.unsupervised.dbcv_calculator import DBCVCalculator
from simba.unsupervised.enums import Clustering, UMLOptions, Unsupervised
from simba.unsupervised.grid_search_visualizers import GridSearchVisualizer
from simba.unsupervised.hdbscan_clusterer import HDBSCANClusterer
from simba.unsupervised.tsne import TSNEGridSearch
from simba.unsupervised.umap_embedder import UmapEmbedder
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty, check_int)
from simba.utils.enums import Formats, Options
from simba.utils.errors import NoSpecifiedOutputError


class GridSearchVisualizerPopUp(PopUpMixin):
    def __init__(self, config_path: str):
        super().__init__(config_path=config_path, title="GRID SEARCH VISUALIZER")

        data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            fg="black",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.data_dir_select = FolderSelect(data_frm, "DATA DIRECTORY:", lblwidth=25)
        self.save_dir_select = FolderSelect(data_frm, "OUTPUT DIRECTORY: ", lblwidth=25)
        data_frm.grid(row=0, column=0, sticky=NW)
        self.data_dir_select.grid(row=0, column=0, sticky=NW)
        self.save_dir_select.grid(row=1, column=0, sticky=NW)

        self.visualization_options = (
            UMLOptions.CATEGORICAL_OPTIONS.value + UMLOptions.CONTINUOUS_OPTIONS.value
        )
        settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=5,
            pady=5,
            fg="black",
        )
        self.scatter_size_dropdown = DropDownMenu(
            settings_frm, "SCATTER SIZE:", UMLOptions.SCATTER_SIZE.value, "25"
        )
        self.scatter_size_dropdown.setChoices(5)
        self.categorical_palette_dropdown = DropDownMenu(
            settings_frm,
            "CATEGORICAL PALETTE:",
            Options.PALETTE_OPTIONS_CATEGORICAL.value,
            "25",
        )
        self.categorical_palette_dropdown.setChoices("Set1")
        self.continuous_palette_dropdown = DropDownMenu(
            settings_frm, "CONTINUOUS PALETTE:", Options.PALETTE_OPTIONS.value, "25"
        )
        self.continuous_palette_dropdown.setChoices("magma")

        settings_frm.grid(row=1, column=0, sticky=NW)
        self.scatter_size_dropdown.grid(row=0, column=0, sticky=NW)
        self.categorical_palette_dropdown.grid(row=1, column=0, sticky=NW)
        self.continuous_palette_dropdown.grid(row=2, column=0, sticky=NW)

        self.define_plots_frm = LabelFrame(
            self.main_frm,
            text="DEFINE PLOTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=5,
            pady=5,
            fg="black",
        )
        self.plot_cnt_dropdown = DropDownMenu(
            self.define_plots_frm,
            "# PLOTS:",
            UMLOptions.GRAPH_CNT.value,
            "25",
            com=lambda x: self.show_plot_table(),
        )
        self.plot_cnt_dropdown.setChoices(UMLOptions.GRAPH_CNT.value[0])
        self.show_plot_table()
        self.define_plots_frm.grid(row=2, column=0, sticky=NW)
        self.plot_cnt_dropdown.grid(row=0, column=0, sticky=NW)

        self.create_run_frm(title="RUN", run_function=self.run)
        self.main_frm.mainloop()

    def show_plot_table(self):
        if hasattr(self, "plot_table"):
            self.plot_table.destroy()

        self.plot_data = {}
        self.plot_table = LabelFrame(
            self.define_plots_frm,
            text="PLOTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=5,
            pady=5,
            fg="black",
        )
        self.scatter_name_header = Label(self.plot_table, text="PLOT NAME").grid(
            row=0, column=0
        )
        self.field_name_header = Label(self.plot_table, text="COLOR VARIABLE").grid(
            row=0, column=1
        )
        for idx in range(int(self.plot_cnt_dropdown.getChoices())):
            row_name = idx
            self.plot_data[row_name] = {}
            self.plot_data[row_name]["label"] = Label(
                self.plot_table, text=f"Scatter {str(idx+1)}:"
            )
            self.plot_data[row_name]["variable"] = DropDownMenu(
                self.plot_table, " ", self.visualization_options, "10", com=None
            )
            self.plot_data[row_name]["variable"].setChoices(
                self.visualization_options[0]
            )
            self.plot_data[idx]["label"].grid(row=idx + 1, column=0, sticky=NW)
            self.plot_data[idx]["variable"].grid(row=idx + 1, column=1, sticky=NW)
        self.plot_table.grid(row=1, column=0, sticky=NW)

    def run(self):
        if len(self.plot_data.keys()) < 1:
            raise NoSpecifiedOutputError(msg="Specify at least one plot")
        settings = {}
        settings["SCATTER_SIZE"] = int(self.scatter_size_dropdown.getChoices())
        settings["CATEGORICAL_PALETTE"] = self.categorical_palette_dropdown.getChoices()
        settings["CONTINUOUS_PALETTE"] = self.continuous_palette_dropdown.getChoices()

        continuous_vars, categorical_vars = [], []
        for k, v in self.plot_data.items():
            if v["variable"].getChoices() in UMLOptions.CONTINUOUS_OPTIONS.value:
                continuous_vars.append(v["variable"].getChoices())
            else:
                categorical_vars.append(v["variable"].getChoices())
        grid_search_visualizer = GridSearchVisualizer(
            model_dir=self.data_dir_select.folder_path,
            save_dir=self.save_dir_select.folder_path,
            settings=settings,
        )
        grid_search_visualizer.continuous_visualizer(continuous_vars=continuous_vars)
        grid_search_visualizer.categorical_visualizer(categoricals=categorical_vars)


# _ = GridSearchVisualizerPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')


class DataExtractorPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="DATA EXTRACTOR")
        ConfigReader.__init__(self, config_path=config_path)
        data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_dir_select = FolderSelect(data_frm, "DATA DIRECTORY:", lblwidth=25)

        data_frm.grid(row=0, column=0, sticky=NW)
        self.data_dir_select.grid(row=0, column=0, sticky=NW)

        data_type_frm = LabelFrame(
            self.main_frm,
            text="DATA TYPE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_type_dropdown = DropDownMenu(
            data_type_frm, "DATA TYPE:", UMLOptions.DATA_TYPES.value, "25"
        )
        self.data_type_dropdown.setChoices(UMLOptions.DATA_TYPES.value[0])
        data_type_frm.grid(row=1, column=0, sticky=NW)
        self.data_type_dropdown.grid(row=0, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run, title="RUN")
        self.main_frm.mainloop()

    def run(self):
        check_if_dir_exists(self.data_dir_select.folder_path)
        DataExtractor(
            config_path=self.config_path,
            data_path=self.data_dir_select.folder_path,
            data_type=self.data_type_dropdown.getChoices(),
        )


# _ = DataExtractorPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')


class FitDimReductionPopUp(PopUpMixin, ConfigReader):
    def __init__(self):
        super().__init__(title="FIT DIMENSIONALITY REDUCTION MODELS")
        self.variance_options = [
            str(x) + "%" for x in UMLOptions.VARIANCE_OPTIONS.value
        ]

        self.dataset_frm = LabelFrame(
            self.main_frm,
            text="DATASET",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.dataset_file_selected = FileSelect(
            self.dataset_frm, "DATASET (PICKLE):", lblwidth=25
        )
        self.dataset_frm.grid(row=0, column=0, sticky=NW)
        self.dataset_file_selected.grid(row=0, column=0, sticky=NW)

        self.save_frm = LabelFrame(
            self.main_frm,
            text="SAVE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.save_dir = FolderSelect(self.save_frm, "SAVE DIRECTORY:", lblwidth=25)
        self.save_frm.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=0, column=0, sticky=NW)

        settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.scaling_dropdown = DropDownMenu(
            settings_frm, "SCALING:", Options.SCALER_NAMES.value, "25"
        )
        self.scaling_dropdown.setChoices(Options.SCALER_NAMES.value[0])
        self.var_threshold_dropdown = DropDownMenu(
            settings_frm, "VARIANCE THRESHOLD:", self.variance_options, "25"
        )
        self.var_threshold_dropdown.setChoices(self.variance_options[0])
        self.algo_dropdown = DropDownMenu(
            settings_frm,
            "ALGORITHM:",
            UMLOptions.DR_ALGO_OPTIONS.value,
            "25",
            com=lambda x: self.show_dr_hyperparameters(),
        )
        self.algo_dropdown.setChoices(UMLOptions.DR_ALGO_OPTIONS.value[0])
        self.show_dr_hyperparameters()

        settings_frm.grid(row=2, column=0, sticky=NW)
        self.scaling_dropdown.grid(row=0, column=0, sticky=NW)
        self.var_threshold_dropdown.grid(row=1, column=0, sticky=NW)
        self.algo_dropdown.grid(row=2, column=0, sticky=NW)
        self.main_frm.mainloop()

    def show_dr_hyperparameters(self):
        if hasattr(self, "hyperparameters_frm"):
            self.hyperparameters_frm.destroy()
            self.value_frm.destroy()
            self.run_frm.destroy()

        self.hyperparameters_frm = LabelFrame(
            self.main_frm,
            text="GRID SEARCH HYPER-PARAMETERS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.value_frm = LabelFrame(self.main_frm, fg="black")
        self.dr_value_entry_box = Entry_Box(self.value_frm, "VALUE: ", "12")
        self.value_frm.grid(row=3, column=0, sticky=NW)
        self.dr_value_entry_box.grid(row=0, column=1, sticky=NW)
        self.hyperparameters_frm.grid(row=4, column=0, sticky=NW)

        if self.algo_dropdown.getChoices() == Unsupervised.UMAP.value:
            Label(self.hyperparameters_frm, text=Unsupervised.N_NEIGHBORS.value).grid(
                row=1, column=0
            )
            Label(self.hyperparameters_frm, text=Unsupervised.MIN_DISTANCE.value).grid(
                row=1, column=1
            )
            Label(self.hyperparameters_frm, text=Unsupervised.SPREAD.value).grid(
                row=1, column=2
            )

            self.n_neighbors_estimators_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.min_distance_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.spread_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )

            neighbours_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.n_neighbors_estimators_listbox,
                    entry_box=self.dr_value_entry_box,
                ),
            )
            min_distance_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.min_distance_listbox,
                    entry_box=self.dr_value_entry_box,
                ),
            )
            spread_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.spread_listbox, entry_box=self.dr_value_entry_box
                ),
            )

            neighbours_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.n_neighbors_estimators_listbox
                ),
            )
            min_distance_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.min_distance_listbox
                ),
            )
            spread_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(list_box=self.spread_listbox),
            )

            self.add_values_to_several_listboxes(
                list_boxes=[
                    self.n_neighbors_estimators_listbox,
                    self.min_distance_listbox,
                    self.spread_listbox,
                ],
                values=[15, 0.1, 1],
            )

            neighbours_add_btn.grid(row=2, column=0)
            min_distance_add_btn.grid(row=2, column=1)
            spread_add_btn.grid(row=2, column=2)
            neighbours_remove_btn.grid(row=3, column=0)
            min_distance_remove_btn.grid(row=3, column=1)
            spread_remove_btn.grid(row=3, column=2)

            self.n_neighbors_estimators_listbox.grid(row=4, column=0, sticky=NW)
            self.min_distance_listbox.grid(row=4, column=1, sticky=NW)
            self.spread_listbox.grid(row=4, column=2, sticky=NW)
            self.create_run_frm(run_function=self.__run_umap_gridsearch)

        elif self.algo_dropdown.getChoices() == Unsupervised.TSNE.value:
            Label(self.hyperparameters_frm, text="PERPLEXITY").grid(row=1, column=0)
            self.perplexity_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            perplexity_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.perplexity_listbox, entry_box=self.dr_value_entry_box
                ),
            )
            perplexity_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.perplexity_listbox
                ),
            )
            perplexity_add_btn.grid(row=2, column=0)
            perplexity_remove_btn.grid(row=3, column=0)
            self.perplexity_listbox.grid(row=4, column=0, sticky=NW)
            self.create_run_frm(run_function=self.__run_tsne_gridsearch)

    def __run_tsne_gridsearch(self):
        self.__get_settings()
        perplexities = [int(x) for x in self.perplexity_listbox.get(0, END)]
        if len(perplexities) == 0:
            raise NoSpecifiedOutputError("Provide value(s) for perplexity")
        hyperparameters = {
            "perplexity": perplexities,
            "scaler": self.scaling_dropdown.getChoices(),
            "variance": self.variance_selected,
        }
        tsne_searcher = TSNEGridSearch(
            data_path=self.data_path, save_dir=self.save_path
        )
        # tsne_searcher.fit(hyperparameters=hyperparameters)

    def __run_umap_gridsearch(self):
        self.__get_settings()
        n_neighbours = [
            float(x) for x in self.n_neighbors_estimators_listbox.get(0, END)
        ]
        min_distances = [float(x) for x in self.min_distance_listbox.get(0, END)]
        spreads = [float(x) for x in self.spread_listbox.get(0, END)]
        if len(min_distances) == 0 or len(n_neighbours) == 0 or len(spreads) == 0:
            raise NoSpecifiedOutputError(
                "Provide at least one hyperparameter value for neighbors, min distances, and spread"
            )
        hyper_parameters = {
            "n_neighbors": n_neighbours,
            "min_distance": min_distances,
            "spread": spreads,
            "scaler": self.scaling_dropdown.getChoices(),
            "variance": self.variance_selected,
        }
        umap_searcher = UmapEmbedder()
        umap_searcher.fit(
            data_path=self.data_path,
            save_dir=self.save_path,
            hyper_parameters=hyper_parameters,
        )

    def __get_settings(self):
        self.variance_selected = (
            int(self.var_threshold_dropdown.getChoices()[:-1]) / 100
        )
        self.save_path = self.save_dir.folder_path
        self.data_path = self.dataset_file_selected.file_path
        self.scaler = self.scaling_dropdown.getChoices()
        check_if_dir_exists(self.save_path)
        check_file_exist_and_readable(self.data_path)


# _ = FitDimReductionPopUp()


class TransformDimReductionPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="DIMENSIONALITY REDUCTION: TRANSFORM")
        self.dim_reduction_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.model_select = FileSelect(
            self.dim_reduction_frm, "MODEL (PICKLE):", lblwidth=25
        )
        self.dataset_select = FileSelect(
            self.dim_reduction_frm, "DATASET (PICKLE):", lblwidth=25
        )
        self.save_dir = FolderSelect(
            self.dim_reduction_frm, "SAVE DIRECTORY: ", lblwidth=25
        )
        self.features_dropdown = DropDownMenu(
            self.dim_reduction_frm,
            "INCLUDE FEATURES:",
            UMLOptions.DATA_FORMATS.value,
            "25",
        )
        self.features_dropdown.setChoices(UMLOptions.DATA_FORMATS.value[0])
        self.save_format_dropdown = DropDownMenu(
            self.dim_reduction_frm, "SAVE FORMATS:", UMLOptions.SAVE_FORMATS.value, "25"
        )
        self.save_format_dropdown.setChoices(UMLOptions.SAVE_FORMATS.value[0])

        self.dim_reduction_frm.grid(row=0, column=0, sticky=NW)
        self.model_select.grid(row=0, column=0, sticky=NW)
        self.dataset_select.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=2, column=0, sticky=NW)
        self.features_dropdown.grid(row=3, column=0, sticky=NW)
        self.save_format_dropdown.grid(row=4, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        check_if_dir_exists(in_dir=self.save_dir.folder_path)
        check_file_exist_and_readable(file_path=self.model_select.file_path)
        check_file_exist_and_readable(file_path=self.dataset_select.file_path)

        umap_searcher = UmapEmbedder()
        umap_searcher.transform(
            data_path=self.dataset_select.file_path,
            model=self.model_select.file_path,
            save_dir=self.save_dir.folder_path,
            settings=None,
        )


# _ = TransformDimReductionPopUp()


class FitClusterModelsPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CLUSTERING FIT: GRID SEARCH")
        self.dataset_frm = LabelFrame(
            self.main_frm,
            text="DATASET",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_dir_selected = FolderSelect(
            self.dataset_frm, "DATA DIRECTORY (PICKLES): ", lblwidth=25
        )
        self.save_frm = LabelFrame(
            self.main_frm,
            text="SAVE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.save_dir = FolderSelect(self.save_frm, "SAVE DIRECTORY: ", lblwidth=25)
        self.algo_frm = LabelFrame(
            self.main_frm,
            text="ALGORITHM",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.algo_dropdown = DropDownMenu(
            self.algo_frm,
            "ALGORITHM:",
            UMLOptions.CLUSTERING_ALGO_OPTIONS.value,
            "25",
            com=lambda x: self.show_hyperparameters(),
        )
        self.algo_dropdown.setChoices(UMLOptions.CLUSTERING_ALGO_OPTIONS.value[0])
        self.value_frm = LabelFrame(self.main_frm, fg="black")
        self.value_entry_box = Entry_Box(self.value_frm, "VALUE:", "25")

        self.dataset_frm.grid(row=0, column=0, sticky=NW)
        self.data_dir_selected.grid(row=0, column=0, sticky=NW)
        self.save_frm.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=0, column=0, sticky=NW)
        self.algo_frm.grid(row=2, column=0, sticky=NW)
        self.algo_dropdown.grid(row=0, column=0, sticky=NW)
        self.value_frm.grid(row=3, column=0, sticky=NW)
        self.value_entry_box.grid(row=0, column=0, sticky=NW)
        self.show_hyperparameters()
        self.main_frm.mainloop()

    def show_hyperparameters(self):
        if hasattr(self, "hyperparameters_frm"):
            self.hyperparameters_frm.destroy()
            self.value_frm.destroy()
            self.run_frm.destroy()
        #
        if self.algo_dropdown.getChoices() == Unsupervised.HDBSCAN.value:
            self.hyperparameters_frm = LabelFrame(
                self.main_frm,
                text="GRID SEARCH CLUSTER HYPER-PARAMETERS",
                font=Formats.LABELFRAME_HEADER_FORMAT.value,
                fg="black",
            )

            Label(self.hyperparameters_frm, text=Clustering.ALPHA.value).grid(
                row=1, column=0
            )
            Label(
                self.hyperparameters_frm, text=Clustering.MIN_CLUSTER_SIZE.value
            ).grid(row=1, column=1)
            Label(self.hyperparameters_frm, text=Clustering.MIN_SAMPLES.value).grid(
                row=1, column=2
            )
            Label(self.hyperparameters_frm, text=Clustering.EPSILON.value).grid(
                row=1, column=3
            )

            self.alpha_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.min_cluster_size_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.min_samples_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.epsilon_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )

            alpha_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.alpha_listbox, entry_box=self.value_entry_box
                ),
            )
            min_cluster_size_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.min_cluster_size_listbox,
                    entry_box=self.value_entry_box,
                ),
            )
            min_samples_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.min_samples_listbox, entry_box=self.value_entry_box
                ),
            )
            epsilon_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.epsilon_listbox, entry_box=self.value_entry_box
                ),
            )

            alpha_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(list_box=self.alpha_listbox),
            )
            min_cluster_size_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.min_cluster_size_listbox
                ),
            )
            min_samples_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.min_samples_listbox
                ),
            )
            epsilon_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(list_box=self.epsilon_listbox),
            )

            self.add_values_to_several_listboxes(
                list_boxes=[
                    self.alpha_listbox,
                    self.min_cluster_size_listbox,
                    self.min_samples_listbox,
                    self.epsilon_listbox,
                ],
                values=[1, 15, 1, 1],
            )

            self.hyperparameters_frm.grid(row=4, column=0, sticky=NW)
            alpha_add_btn.grid(row=2, column=0)
            min_cluster_size_add_btn.grid(row=2, column=1)
            min_samples_add_btn.grid(row=2, column=2)
            epsilon_add_btn.grid(row=2, column=3)

            alpha_remove_btn.grid(row=3, column=0)
            min_cluster_size_remove_btn.grid(row=3, column=1)
            min_samples_remove_btn.grid(row=3, column=2)
            epsilon_remove_btn.grid(row=3, column=3)

            self.alpha_listbox.grid(row=4, column=0, sticky=NW)
            self.min_cluster_size_listbox.grid(row=4, column=1, sticky=NW)
            self.min_samples_listbox.grid(row=4, column=2, sticky=NW)
            self.epsilon_listbox.grid(row=4, column=3, sticky=NW)

            self.create_run_frm(run_function=self.run_hdbscan_clustering)

    def __get_settings(self):
        self.data_directory = self.data_dir_selected.folder_path
        self.save_directory = self.save_dir.folder_path
        check_if_dir_exists(self.data_dir_selected.folder_path)
        check_if_dir_exists(self.save_dir.folder_path)

    def run_hdbscan_clustering(self):
        self.__get_settings()
        alphas = [float(x) for x in self.alpha_listbox.get(0, END)]
        min_cluster_sizes = [int(x) for x in self.min_cluster_size_listbox.get(0, END)]
        min_samples = [int(x) for x in self.min_samples_listbox.get(0, END)]
        epsilons = [float(x) for x in self.epsilon_listbox.get(0, END)]
        hyper_parameters = {
            "alpha": alphas,
            "min_cluster_size": min_cluster_sizes,
            "min_samples": min_samples,
            "cluster_selection_epsilon": epsilons,
        }
        clusterer = HDBSCANClusterer()
        clusterer.fit(
            data_path=self.data_directory,
            save_dir=self.save_directory,
            hyper_parameters=hyper_parameters,
        )


# _ = FitClusterModelsPopUp()


class TransformClustererPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CLUSTERING: TRANSFORM")
        self.settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.model_select = FileSelect(
            self.settings_frm, "CLUSTER MODEL (PICKLE):", lblwidth=25
        )
        self.data_select = FileSelect(
            self.settings_frm, "DATASET (PICKLE):", lblwidth=25
        )
        self.save_dir = FolderSelect(self.settings_frm, "SAVE DIRECTORY:", lblwidth=25)
        self.features_dropdown = DropDownMenu(
            self.settings_frm, "INCLUDE FEATURES:", UMLOptions.DATA_FORMATS.value, "25"
        )
        self.features_dropdown.setChoices(UMLOptions.DATA_FORMATS.value[0])
        self.save_format_dropdown = DropDownMenu(
            self.settings_frm, "SAVE FORMATS:", UMLOptions.SAVE_FORMATS.value, "25"
        )
        self.save_format_dropdown.setChoices(UMLOptions.SAVE_FORMATS.value[0])
        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.model_select.grid(row=0, column=0, sticky=NW)
        self.data_select.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=2, column=0, sticky=NW)
        self.features_dropdown.grid(row=3, column=0, sticky=NW)
        self.save_format_dropdown.grid(row=4, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(self.model_select.file_path)
        check_file_exist_and_readable(self.data_select.file_path)
        check_if_dir_exists(self.save_dir.folder_path)
        settings = {
            Unsupervised.DATA.value: self.features_dropdown.getChoices(),
            Unsupervised.FORMAT.value: self.save_format_dropdown.getChoices(),
        }

        clusterer = HDBSCANClusterer()
        clusterer.transform(
            data_path=self.data_select.file_path,
            model=self.model_select.file_path,
            save_dir=self.save_dir.folder_path,
            settings=settings,
        )


# _ = TransformClustererPopUp()


class ClusterVisualizerPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="CLUSTER VIDEO VISUALIZATIONS")
        ConfigReader.__init__(self, config_path=config_path)
        self.include_pose_var = BooleanVar(value=False)

        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.videos_dir_select = FolderSelect(
            self.data_frm, "VIDEOS DIRECTORY:", lblwidth=25
        )
        self.dataset_file_selected = FileSelect(
            self.data_frm, "DATASET (PICKLE): ", lblwidth=25
        )
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.videos_dir_select.grid(row=0, column=0, sticky=NW)

        self.settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.include_pose = Checkbutton(
            self.settings_frm,
            text="INCLUDE POSE-ESTIMATION",
            variable=self.include_pose_var,
            command=lambda: self.enable_entrybox_from_checkbox(
                check_box_var=self.include_pose_var,
                entry_boxes=[self.circle_size_entry],
            ),
        )
        self.circle_size_entry = Entry_Box(
            self.settings_frm, "CIRCLE SIZE: ", "25", validation="numeric"
        )
        self.circle_size_entry.entry_set(val=5)
        self.circle_size_entry.set_state(setstatus="disable")

        self.speed_dropdown = DropDownMenu(
            self.settings_frm, "VIDEO SPEED:", UMLOptions.SPEED_OPTIONS.value, "25"
        )
        self.speed_dropdown.setChoices(1.0)
        self.settings_frm.grid(row=1, column=0, sticky=NW)
        self.include_pose.grid(row=0, column=0, sticky=NW)
        self.circle_size_entry.grid(row=1, column=0, sticky=NW)
        self.speed_dropdown.grid(row=2, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(file_path=self.dataset_file_selected.file_path)
        check_if_dir_exists(in_dir=self.videos_dir_select.folder_path)
        speed = float(self.speed_dropdown.getChoices())
        if self.include_pose_var.get():
            check_int(name="CIRCLE SIZE", value=self.circle_size_entry.entry_get)
            circle_size = self.circle_size_entry.entry_get
        else:
            circle_size = np.inf
        settings = {
            "videos_speed": speed,
            "pose": {
                "include": self.include_pose_var.get(),
                "circle_size": circle_size,
            },
        }
        cluster_visualizer = ClusterVisualizer(
            config_path=self.config_path,
            settings=settings,
            video_dir=self.videos_dir_select.folder_path,
            data_path=self.dataset_file_selected.file_path,
        )
        cluster_visualizer.run()


# _ = ClusterVisualizerPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')
#
class ClusterFrequentistStatisticsPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="CLUSTER FREQUENTIST STATISTICS")
        ConfigReader.__init__(self, config_path=config_path)
        self.descriptive_stats_var = BooleanVar(value=True)
        self.oneway_anova_var = BooleanVar(value=True)
        self.tukey_var = BooleanVar(value=True)
        self.use_scaled_var = BooleanVar(value=False)

        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.model_select = FileSelect(self.data_frm, "CLUSTERER PATH:", lblwidth=25)
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.model_select.grid(row=0, column=0, sticky=NW)

        self.stats_frm = LabelFrame(
            self.main_frm,
            text="STATISTICS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.descriptive_stats_cb = Checkbutton(
            self.stats_frm,
            text="CLUSTER DESCRIPTIVE STATISTICS",
            variable=self.descriptive_stats_var,
        )
        self.oneway_anova_cb = Checkbutton(
            self.stats_frm,
            text="CLUSTER FEATURE ONE-WAY ANOVA",
            variable=self.oneway_anova_var,
        )
        self.feature_tukey_posthoc_cb = Checkbutton(
            self.stats_frm,
            text="CLUSTER FEATURE POST-HOC (TUKEY)",
            variable=self.tukey_var,
        )
        self.use_scaled_cb = Checkbutton(
            self.stats_frm,
            text="USE SCALED FEATURE VALUES",
            variable=self.use_scaled_var,
        )

        self.stats_frm.grid(row=1, column=0, sticky=NW)
        self.descriptive_stats_cb.grid(row=0, column=0, sticky=NW)
        self.oneway_anova_cb.grid(row=1, column=0, sticky=NW)
        self.feature_tukey_posthoc_cb.grid(row=2, column=0, sticky=NW)
        self.use_scaled_cb.grid(row=3, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(self.model_select.file_path)
        settings = {
            "scaled": self.use_scaled_var.get(),
            "anova": self.oneway_anova_var.get(),
            "tukey_posthoc": self.tukey_var.get(),
            "descriptive_statistics": self.descriptive_stats_var.get(),
        }

        calculator = ClusterFrequentistCalculator(
            config_path=self.config_path,
            data_path=self.model_select.file_path,
            settings=settings,
        )
        calculator.run()


# _ = ClusterFrequentistStatisticsPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')


class ClusterXAIPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        super().__init__(title="CLUSTER XAI STATISTICS")
        ConfigReader.__init__(self, config_path=config_path)
        self.gini_importance_var = BooleanVar(value=True)
        self.permutation_importance_var = BooleanVar(value=True)
        self.shap_var = BooleanVar(value=False)

        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.model_select = FileSelect(self.data_frm, "MODEL PATH:", lblwidth=25)
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.model_select.grid(row=0, column=0, sticky=NW)

        self.settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.gini_importance_cb = Checkbutton(
            self.settings_frm,
            text="CLUSTER RF GINI IMPORTANCE",
            variable=self.gini_importance_var,
        )
        self.permutation_cb = Checkbutton(
            self.settings_frm,
            text="CLUSTER RF PERMUTATION IMPORTANCE",
            variable=self.permutation_importance_var,
        )
        self.shap_method_dropdown = DropDownMenu(
            self.settings_frm,
            "SHAP METHOD:",
            UMLOptions.SHAP_CLUSTER_METHODS.value,
            "25",
        )
        self.shap_method_dropdown.setChoices(UMLOptions.SHAP_CLUSTER_METHODS.value[0])
        self.shap_sample_dropdown = DropDownMenu(
            self.settings_frm,
            "SHAP SAMPLES:",
            UMLOptions.SHAP_SAMPLE_OPTIONS.value,
            "25",
        )
        self.shap_sample_dropdown.setChoices(100)
        self.shap_method_dropdown.disable()
        self.shap_sample_dropdown.disable()
        self.shap_cb = Checkbutton(
            self.settings_frm,
            text="CLUSTER RF SHAP VALUES",
            variable=self.shap_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.shap_var,
                dropdown_menus=[self.shap_method_dropdown, self.shap_sample_dropdown],
            ),
        )

        self.settings_frm.grid(row=1, column=0, sticky=NW)
        self.gini_importance_cb.grid(row=0, column=0, sticky=NW)
        self.permutation_cb.grid(row=1, column=0, sticky=NW)
        self.shap_cb.grid(row=2, column=0, sticky=NW)
        self.shap_method_dropdown.grid(row=3, column=0, sticky=NW)
        self.shap_sample_dropdown.grid(row=4, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(file_path=self.model_select.file_path)
        settings = {
            "gini_importance": self.gini_importance_var.get(),
            "permutation_importance": self.permutation_importance_var.get(),
            "shap": {
                "method": self.shap_method_dropdown.getChoices(),
                "run": self.shap_var.get(),
                "sample": self.shap_sample_dropdown.getChoices(),
            },
        }
        xai_calculator = ClusterXAICalculator(
            data_path=self.model_select.file_path,
            settings=settings,
            config_path=self.config_path,
        )
        xai_calculator.run()


# _ = ClusterXAIPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')


class EmbedderCorrelationsPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="EMBEDDING CORRELATIONS")
        ConfigReader.__init__(self, config_path=config_path)
        self.spearman_var = BooleanVar(value=True)
        self.pearsons_var = BooleanVar(value=True)
        self.kendall_var = BooleanVar(value=True)
        self.plots_var = BooleanVar(value=False)

        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_file_selected = FileSelect(self.data_frm, "DATASET (PICKLE):")
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.data_file_selected.grid(row=0, column=0, sticky=NW)

        self.settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.spearman_cb = Checkbutton(
            self.settings_frm, text="SPEARMAN", variable=self.spearman_var
        )
        self.pearsons_cb = Checkbutton(
            self.settings_frm, text="PEARSONS", variable=self.pearsons_var
        )
        self.kendall_cb = Checkbutton(
            self.settings_frm, text="KENDALL", variable=self.kendall_var
        )
        self.plot_correlation_dropdown = DropDownMenu(
            self.settings_frm,
            "PLOT CORRELATION:",
            UMLOptions.CORRELATION_OPTIONS.value,
            "25",
        )
        self.plot_correlation_clr_dropdown = DropDownMenu(
            self.settings_frm, "PLOT CORRELATION:", Options.PALETTE_OPTIONS.value, "25"
        )
        self.plot_correlation_dropdown.setChoices(
            UMLOptions.CORRELATION_OPTIONS.value[0]
        )
        self.plot_correlation_dropdown.disable()
        self.plot_correlation_clr_dropdown.setChoices(Options.PALETTE_OPTIONS.value[0])
        self.plot_correlation_clr_dropdown.disable()
        self.plots_cb = Checkbutton(
            self.settings_frm,
            text="PLOTS",
            variable=self.plots_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.plots_var,
                dropdown_menus=[
                    self.plot_correlation_dropdown,
                    self.plot_correlation_clr_dropdown,
                ],
            ),
        )

        self.settings_frm.grid(row=1, column=0, sticky=NW)
        self.spearman_cb.grid(row=0, column=0, sticky=NW)
        self.pearsons_cb.grid(row=1, column=0, sticky=NW)
        self.kendall_cb.grid(row=2, column=0, sticky=NW)
        self.plots_cb.grid(row=3, column=0, sticky=NW)
        self.plot_correlation_dropdown.grid(row=4, column=0, sticky=NW)
        self.plot_correlation_clr_dropdown.grid(row=5, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(self.data_file_selected.file_path)
        settings = {
            "correlations": [],
            "plots": {"create": False, "correlations": None, "palette": None},
        }
        if self.spearman_var.get():
            settings["correlations"].append("spearman")
        if self.pearsons_var.get():
            settings["correlations"].append("pearson")
        if self.kendall_var.get():
            settings["correlations"].append("kendall")

        calculator = EmbeddingCorrelationCalculator(
            config_path=self.config_path,
            data_path=self.data_file_selected.file_path,
            settings=settings,
        )
        calculator.run()


# _ = EmbedderCorrelationsPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')


class PrintEmBeddingInfoPopUp(PopUpMixin, ConfigReader, UnsupervisedMixin):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="PRINT EMBEDDING MODEL INFO")
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.dataset_file_selected = FileSelect(
            self.data_frm, "DATASET (PICKLE): ", lblwidth=25
        )
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.dataset_file_selected.grid(row=0, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(file_path=self.dataset_file_selected.file_path)
        data = self.read_pickle(data_path=self.dataset_file_selected.file_path)
        parameters = {
            **data[Unsupervised.DR_MODEL.value][Unsupervised.PARAMETERS.value]
        }
        print(parameters)


# _ = PrintEmBeddingInfoPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')


class DBCVPopUp(PopUpMixin, ConfigReader, UnsupervisedMixin):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="DENSITY BASED CLUSTER VALIDATION")
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)

        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.folder_selected = FolderSelect(
            self.data_frm, "DATASETS (DIRECTORY WITH PICKLES):", lblwidth=35
        )
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.folder_selected.grid(row=0, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        check_if_dir_exists(in_dir=self.folder_selected.folder_path)
        data_paths = glob.glob(self.folder_selected.folder_path + "/*.pickle")
        check_if_filepath_list_is_empty(
            filepaths=data_paths,
            error_msg=f"No pickle files in {self.folder_selected.folder_path}",
        )
        dbcv_calculator = DBCVCalculator(
            data_path=self.folder_selected.folder_path, config_path=self.config_path
        )
        dbcv_calculator.run()


# _ = DBCVPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os
from collections import defaultdict
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_clf_calculator import ROIClfCalculator
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.enums import Keys, Links
from simba.utils.errors import (NoChoosenClassifierError,
                                NoChoosenMeasurementError, NoChoosenROIError,
                                ROICoordinatesNotFoundError)


class ClfByROIPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(
                expected_file_path=self.roi_coordinates_path,
                source=self.__class__.__name__,
            )
        PopUpMixin.__init__(self, title="CLASSIFICATIONS BY ROI")
        self.read_roi_data()
        body_part_menu = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Select body part",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ANALYZE_ML_RESULTS.value,
        )
        ROI_menu = LabelFrame(self.main_frm, text="Select ROI(s)", padx=5, pady=5)
        classifier_menu = LabelFrame(
            self.main_frm, text="Select classifier(s)", padx=5, pady=5
        )
        measurements_menu = LabelFrame(
            self.main_frm, text="Select measurements", padx=5, pady=5
        )
        self.total_time_var = BooleanVar()
        self.start_bouts_var = BooleanVar()
        self.end_bouts_var = BooleanVar()
        self.total_time_cb = Checkbutton(
            measurements_menu,
            text="Total time by ROI (s)",
            variable=self.total_time_var,
        )
        self.start_bouts_cb = Checkbutton(
            measurements_menu,
            text="Started bouts by ROI (count)",
            variable=self.start_bouts_var,
        )
        self.end_bouts_cb = Checkbutton(
            measurements_menu,
            text="Ended bouts by ROI (count)",
            variable=self.end_bouts_var,
        )
        self.ROI_check_boxes_status_dict = {}
        self.clf_check_boxes_status_dict = {}

        for row_number, ROI in enumerate(self.roi_types_names_lst):
            self.ROI_check_boxes_status_dict[ROI] = IntVar()
            ROI_check_button = Checkbutton(
                ROI_menu, text=ROI, variable=self.ROI_check_boxes_status_dict[ROI]
            )
            ROI_check_button.grid(row=row_number, sticky=W)

        for row_number, clf_name in enumerate(self.clf_names):
            self.clf_check_boxes_status_dict[clf_name] = IntVar()
            clf_check_button = Checkbutton(
                classifier_menu,
                text=clf_name,
                variable=self.clf_check_boxes_status_dict[clf_name],
            )
            clf_check_button.grid(row=row_number, sticky=W)

        self.choose_bp = DropDownMenu(
            body_part_menu, "Body part", self.body_parts_lst, "12"
        )
        self.choose_bp.setChoices(self.body_parts_lst[0])
        self.choose_bp.grid(row=0, sticky=W)
        run_analysis_button = Button(
            self.main_frm,
            text="Analyze classifications in each ROI",
            command=lambda: self.run_clf_by_ROI_analysis(),
        )
        body_part_menu.grid(row=0, sticky=W, padx=10, pady=10)
        ROI_menu.grid(row=1, sticky=W, padx=10, pady=10)
        classifier_menu.grid(row=2, sticky=W, padx=10, pady=10)
        self.total_time_cb.grid(row=0, sticky=NW)
        self.start_bouts_cb.grid(row=1, sticky=NW)
        self.end_bouts_cb.grid(row=2, sticky=NW)
        measurements_menu.grid(row=3, sticky=W, padx=10, pady=10)
        run_analysis_button.grid(row=4, sticky=W, padx=10, pady=10)

        self.main_frm.mainloop()

    def run_clf_by_ROI_analysis(self):
        body_part_list = [self.choose_bp.getChoices()]
        ROI_dict_lists, behavior_list = defaultdict(list), []
        measurements_list = []
        for loop_val, ROI_entry in enumerate(self.ROI_check_boxes_status_dict):
            check_val = self.ROI_check_boxes_status_dict[ROI_entry]
            if check_val.get() == 1:
                shape_type = (
                    self.roi_types_names_lst[loop_val].split(":")[0].replace(":", "")
                )
                shape_name = self.roi_types_names_lst[loop_val].split(":")[1][1:]
                ROI_dict_lists[shape_type].append(shape_name)

        for measurement_var, measurement_name in zip(
            [
                self.total_time_var.get(),
                self.start_bouts_var.get(),
                self.end_bouts_var.get(),
            ],
            [
                "Total time by ROI (s)",
                "Started bouts by ROI (count)",
                "Ended bouts by ROI (count)",
            ],
        ):
            if measurement_var:
                measurements_list.append(measurement_name)

        for loop_val, clf_entry in enumerate(self.clf_check_boxes_status_dict):
            check_val = self.clf_check_boxes_status_dict[clf_entry]
            if check_val.get() == 1:
                behavior_list.append(self.clf_names[loop_val])
        if len(ROI_dict_lists) == 0:
            raise NoChoosenROIError(source=self.__class__.__name__)
        if len(behavior_list) == 0:
            raise NoChoosenClassifierError(source=self.__class__.__name__)
        if len(measurements_list) == 0:
            raise NoChoosenMeasurementError(source=self.__class__.__name__)
        else:
            self.clf_roi_analyzer = ROIClfCalculator(config_ini=self.config_path)
            self.clf_roi_analyzer.run(
                ROI_dict_lists=ROI_dict_lists,
                behavior_list=behavior_list,
                body_part_list=body_part_list,
                measurements=measurements_list,
            )


#
# _ = ClfByROIPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform
from copy import deepcopy
from typing import Dict, List, Optional, Union

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int)
from simba.utils.data import create_color_palette
from simba.utils.enums import ConfigKey, Dtypes, Formats, TagNames
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data, read_config_entry,
                                    read_df)


def _multiprocess_sklearn_video(
    data: np.array,
    video_path: str,
    video_save_dir: str,
    frame_save_dir: str,
    clf_colors: list,
    models_info: dict,
    bp_dict: dict,
    text_attr: dict,
    rotate: bool,
    print_timers: bool,
    video_setting: bool,
    frame_setting: bool,
    pose_threshold: float,
):
    fourcc, font = cv2.VideoWriter_fourcc(*"mp4v"), cv2.FONT_HERSHEY_COMPLEX
    video_meta_data = get_video_meta_data(video_path=video_path)
    cap = cv2.VideoCapture(video_path)
    group = data["group"].iloc[0]
    start_frm, current_frm, end_frm = (
        data["index"].iloc[0],
        data["index"].iloc[0],
        data["index"].iloc[-1],
    )

    if video_setting:
        video_save_path = os.path.join(video_save_dir, "{}.mp4".format(str(group)))
        video_writer = cv2.VideoWriter(
            video_save_path,
            fourcc,
            video_meta_data["fps"],
            (video_meta_data["width"], video_meta_data["height"]),
        )
    cap.set(1, start_frm)
    while current_frm < end_frm:
        ret, img = cap.read()
        add_spacer = 2
        for animal_name, animal_data in bp_dict.items():
            animal_clr = animal_data["colors"]
            id_flag_cords = None
            for bp_no in range(len(animal_data["X_bps"])):
                bp_clr = animal_clr[bp_no]
                x_bp, y_bp, p_bp = (
                    animal_data["X_bps"][bp_no],
                    animal_data["Y_bps"][bp_no],
                    animal_data["P_bps"][bp_no],
                )
                bp_cords = data.loc[current_frm, [x_bp, y_bp, p_bp]]
                if bp_cords[p_bp] > pose_threshold:
                    cv2.circle(
                        img,
                        (int(bp_cords[x_bp]), int(bp_cords[y_bp])),
                        0,
                        bp_clr,
                        text_attr["circle_scale"],
                    )
                    if ("centroid" in x_bp.lower()) or ("center" in x_bp.lower()):
                        id_flag_cords = (int(bp_cords[x_bp]), int(bp_cords[y_bp]))

            if not id_flag_cords:
                id_flag_cords = (int(bp_cords[x_bp]), int(bp_cords[y_bp]))
            cv2.putText(
                img,
                animal_name,
                id_flag_cords,
                font,
                text_attr["font_size"],
                animal_clr[0],
                text_attr["text_thickness"],
            )
        if rotate:
            img = cv2.rotate(img, cv2.ROTATE_90_CLOCKWISE)
        if print_timers:
            cv2.putText(
                img,
                "Timers",
                (10, text_attr["spacing_scale"]),
                font,
                text_attr["font_size"],
                (0, 255, 0),
                text_attr["text_thickness"],
            )
        frame_results = {}
        for model in models_info.values():
            frame_results[model["model_name"]] = data.loc[
                current_frm, model["model_name"]
            ]
            if print_timers:
                cumulative_time = round(
                    data.loc[current_frm, model["model_name"] + "_cumsum"]
                    / video_meta_data["fps"],
                    3,
                )
                cv2.putText(
                    img,
                    "{} {}s".format(model["model_name"], str(cumulative_time)),
                    (10, text_attr["spacing_scale"] * add_spacer),
                    font,
                    text_attr["font_size"],
                    (255, 0, 0),
                    text_attr["text_thickness"],
                )
                add_spacer += 1

        cv2.putText(
            img,
            "Ensemble prediction",
            (10, int(text_attr["spacing_scale"] * add_spacer)),
            font,
            text_attr["font_size"],
            (0, 255, 0),
            text_attr["text_thickness"],
        )
        add_spacer += 1
        for clf_cnt, (clf_name, clf_results) in enumerate(frame_results.items()):
            if clf_results == 1:
                cv2.putText(
                    img,
                    clf_name,
                    (10, int(text_attr["spacing_scale"] * add_spacer)),
                    font,
                    text_attr["font_size"],
                    clf_colors[clf_cnt],
                    text_attr["text_thickness"],
                )
                add_spacer += 1
        if video_setting:
            video_writer.write(img)
        if frame_setting:
            frame_save_name = os.path.join(
                frame_save_dir, "{}.png".format(str(current_frm))
            )
            cv2.imwrite(frame_save_name, img)
        current_frm += 1
        print(
            "Multi-processing video frame {} on core {}...".format(
                str(current_frm), str(group)
            )
        )

    cap.release()
    if video_setting:
        video_writer.release()

    return group


class PlotSklearnResultsMultiProcess(ConfigReader, TrainModelMixin, PlottingMixin):
    """
    Plot classification results on videos. Results are stored in the
    `project_folder/frames/output/sklearn_results` directory of the SimBA project.

    .. seelalso::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-10-sklearn-visualization__.
        For non-multiptocess class, see :meth:`simba.plotting.plot_clf_results.PlotSklearnResultsSingleCore`.

    .. image:: _static/img/sklearn_visualization.gif
       :width: 600
       :align: center


    :param str config_path: path to SimBA project config file in Configparser format
    :param bool rotate: If True, the output video will be rotated 90 degrees from the input.
    :param bool video_setting: If True, SimBA will create compressed videos.
    :param bool frame_setting: If True, SimBA will create individual frames.
    :param str video_file_path: path to video file to create classification visualizations for.
    :param int cores: Number of cores to use


    Examples
    ----------
    >>> clf_plotter = PlotSklearnResultsMultiProcess(config_path='MyProjectConfig', video_setting=True, frame_setting=False, rotate=False, video_file_path='VideoPath', cores=5)
    >>> clf_plotter.run()
    """

    def __init__(
        self,
        config_path: str,
        video_setting: bool,
        frame_setting: bool,
        cores: int,
        video_file_path: Optional[str] = None,
        text_settings: Optional[Union[Dict[str, float], bool]] = False,
        rotate: Optional[bool] = False,
        print_timers: Optional[bool] = True,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        PlottingMixin.__init__(self)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )

        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

        if (not video_setting) and (not frame_setting):
            raise NoSpecifiedOutputError(
                msg="SIMBA ERROR: Please choose to create a video and/or frames. SimBA found that you ticked neither video and/or frames",
                source=self.__class__.__name__,
            )
        self.video_file_path, self.print_timers, self.text_settings = (
            video_file_path,
            print_timers,
            text_settings,
        )
        self.video_setting, self.frame_setting, self.cores, self.rotate = (
            video_setting,
            frame_setting,
            cores,
            rotate,
        )
        if video_file_path is not None:
            check_file_exist_and_readable(os.path.join(self.video_dir, video_file_path))
        if not os.path.exists(self.sklearn_plot_dir):
            os.makedirs(self.sklearn_plot_dir)
        self.pose_threshold = read_config_entry(
            self.config,
            ConfigKey.THRESHOLD_SETTINGS.value,
            ConfigKey.SKLEARN_BP_PROB_THRESH.value,
            Dtypes.FLOAT.value,
            0.00,
        )
        self.model_dict = self.get_model_info(self.config, self.clf_cnt)
        self.clf_colors = create_color_palette(
            pallete_name="Set1", increments=self.clf_cnt
        )
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)

    def __get_print_settings(self):
        self.text_attr = {}
        if self.text_settings is False:
            self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
                60,
                12,
                1500,
                1.1,
            )
            self.max_dim = max(
                self.video_meta_data["width"], self.video_meta_data["height"]
            )
            self.text_attr["circle_scale"] = int(
                self.radius_scale / (self.res_scale / self.max_dim)
            )
            self.text_attr["font_size"] = float(
                self.font_scale / (self.res_scale / self.max_dim)
            )
            self.text_attr["spacing_scale"] = int(
                self.space_scale / (self.res_scale / self.max_dim)
            )
            self.text_attr["text_thickness"] = 2
        else:
            check_float(name="ERROR: TEXT SIZE", value=self.text_settings["font_size"])
            check_int(name="ERROR: SPACE SIZE", value=self.text_settings["space_size"])
            check_int(
                name="ERROR: TEXT THICKNESS", value=self.text_settings["text_thickness"]
            )
            check_int(
                name="ERROR: CIRCLE SIZE", value=self.text_settings["circle_size"]
            )
            self.text_attr["font_size"] = float(self.text_settings["font_size"])
            self.text_attr["spacing_scale"] = int(self.text_settings["space_size"])
            self.text_attr["text_thickness"] = int(self.text_settings["text_thickness"])
            self.text_attr["circle_scale"] = int(self.text_settings["circle_size"])

    def __index_df_for_multiprocessing(self, data: list) -> list:
        for cnt, df in enumerate(data):
            df["group"] = cnt
        return data

    def create_visualizations(self):
        video_timer = SimbaTimer(start=True)
        _, self.video_name, _ = get_fn_ext(self.file_path)
        self.data_df = read_df(self.file_path, self.file_type).reset_index(drop=True)
        self.video_settings, _, self.fps = self.read_video_info(
            video_name=self.video_name
        )
        self.video_path = self.find_video_of_file(self.video_dir, self.video_name)
        self.video_meta_data = get_video_meta_data(self.video_path)
        height, width = deepcopy(self.video_meta_data["height"]), deepcopy(
            self.video_meta_data["width"]
        )
        self.video_frame_dir, self.video_temp_dir = None, None
        if self.frame_setting:
            self.video_frame_dir = os.path.join(self.sklearn_plot_dir, self.video_name)
            if not os.path.exists(self.video_frame_dir):
                os.makedirs(self.video_frame_dir)
        if self.video_setting:
            self.video_save_path = os.path.join(
                self.sklearn_plot_dir, self.video_name + ".mp4"
            )
            self.video_temp_dir = os.path.join(
                self.sklearn_plot_dir, self.video_name, "temp"
            )
            if not os.path.exists(self.video_temp_dir):
                os.makedirs(self.video_temp_dir)
        if self.rotate:
            self.video_meta_data["height"], self.video_meta_data["width"] = (
                width,
                height,
            )
        self.__get_print_settings()

        for model in self.model_dict.values():
            self.data_df[model["model_name"] + "_cumsum"] = self.data_df[
                model["model_name"]
            ].cumsum()
        self.data_df["index"] = self.data_df.index
        data = np.array_split(self.data_df, self.cores)
        frm_per_core = data[0].shape[0]

        data = self.__index_df_for_multiprocessing(data=data)
        with multiprocessing.Pool(
            self.cores, maxtasksperchild=self.maxtasksperchild
        ) as pool:
            constants = functools.partial(
                _multiprocess_sklearn_video,
                clf_colors=self.clf_colors,
                bp_dict=self.animal_bp_dict,
                video_save_dir=self.video_temp_dir,
                frame_save_dir=self.video_frame_dir,
                models_info=self.model_dict,
                text_attr=self.text_attr,
                rotate=self.rotate,
                video_path=self.video_path,
                print_timers=self.print_timers,
                video_setting=self.video_setting,
                frame_setting=self.frame_setting,
                pose_threshold=self.pose_threshold,
            )

            for cnt, result in enumerate(
                pool.imap(constants, data, chunksize=self.multiprocess_chunksize)
            ):
                print(
                    "Image {}/{}, Video {}/{}...".format(
                        str(int(frm_per_core * (result + 1))),
                        str(len(self.data_df)),
                        str(self.file_cnt + 1),
                        str(len(self.files_found)),
                    )
                )
            if self.video_setting:
                print("Joining {} multiprocessed video...".format(self.video_name))
                concatenate_videos_in_folder(
                    in_folder=self.video_temp_dir, save_path=self.video_save_path
                )
            video_timer.stop_timer()
            pool.terminate()
            pool.join()
            print(
                "Video {} complete (elapsed time: {}s)...".format(
                    self.video_name, video_timer.elapsed_time_str
                )
            )

    def run(self):
        if self.video_file_path is None:
            self.files_found = self.machine_results_paths
            print(
                "Processing {} videos...".format(str(len(self.machine_results_paths)))
            )
            for file_cnt, file_path in enumerate(self.machine_results_paths):
                self.file_cnt, self.file_path = file_cnt, file_path
                self.create_visualizations()
        else:
            print("Processing 1 video...")
            self.file_cnt, file_path = 0, self.video_file_path
            _, file_name, _ = get_fn_ext(file_path)
            self.file_path = os.path.join(
                self.machine_results_dir, file_name + "." + self.file_type
            )
            self.files_found = [self.file_path]
            check_file_exist_and_readable(self.file_path)
            self.create_visualizations()

        self.timer.stop_timer()
        if self.video_setting:
            stdout_success(
                msg=f"{len(self.files_found)} videos saved in project_folder/frames/output/sklearn_results directory",
                elapsed_time=self.timer.elapsed_time_str,
                source=self.__class__.__name__,
            )
        if self.frame_setting:
            stdout_success(
                f"Frames for {len(self.files_found)} videos saved in sub-folders within project_folder/frames/output/sklearn_results directory",
                elapsed_time=self.timer.elapsed_time_str,
                source=self.__class__.__name__,
            )


# clf_plotter = PlotSklearnResultsMultiProcess(config_path='/Users/simon/Desktop/envs/troubleshooting/DLC_2_Black_animals/project_folder/project_config.ini', video_setting=True, frame_setting=False, rotate=False, video_file_path='Together_1.avi', cores=5)
# clf_plotter.run()



__author__ = "Simon Nilsson"

import os
from tkinter import *

from simba.data_processors.interpolation_smoothing import Interpolate, Smooth
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import DropDownMenu, Entry_Box, FolderSelect
from simba.utils.checks import check_int
from simba.utils.enums import Options
from simba.utils.errors import NotDirectoryError


class InterpolatePopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="INTERPOLATE POSE")
        ConfigReader.__init__(self, config_path=config_path)
        self.input_dir = FolderSelect(self.main_frm, "DATA DIRECTORY:", lblwidth=25)
        self.method_dropdown = DropDownMenu(
            self.main_frm,
            "INTERPOLATION METHOD:",
            Options.INTERPOLATION_OPTIONS.value,
            "25",
        )
        self.method_dropdown.setChoices(Options.INTERPOLATION_OPTIONS.value[0])
        run_btn = Button(
            self.main_frm,
            text="RUN INTERPOLATION",
            fg="blue",
            command=lambda: self.run(),
        )
        self.input_dir.grid(row=0, column=0, sticky=NW)
        self.method_dropdown.grid(row=1, column=0, sticky=NW)
        run_btn.grid(row=2, column=0, sticky=NW)

    def run(self):
        if not os.path.isdir(self.input_dir.folder_path):
            raise NotDirectoryError(
                msg=f"{self.input_dir.folder_path} is not a valid directory.",
                source=self.__class__.__name__,
            )
        Interpolate(
            config_path=self.config_path,
            method=self.method_dropdown.getChoices(),
            input_path=self.input_dir.folder_path,
        )


class SmoothingPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="SMOOTH POSE")
        ConfigReader.__init__(self, config_path=config_path)
        self.input_dir = FolderSelect(self.main_frm, "DATA DIRECTORY:", lblwidth=20)
        self.time_window = Entry_Box(
            self.main_frm, "TIME WINDOW (MS):", "20", validation="numeric"
        )
        self.method_dropdown = DropDownMenu(
            self.main_frm, "SMOOTHING METHOD:", Options.SMOOTHING_OPTIONS.value, "20"
        )
        self.method_dropdown.setChoices(Options.SMOOTHING_OPTIONS.value[0])
        run_btn = Button(
            self.main_frm, text="RUN SMOOTHING", fg="blue", command=lambda: self.run()
        )

        self.input_dir.grid(row=0, column=0, sticky=NW)
        self.method_dropdown.grid(row=1, column=0, sticky=NW)
        self.time_window.grid(row=2, column=0, sticky=NW)
        run_btn.grid(row=3, column=0, sticky=NW)

    def run(self):
        if not os.path.isdir(self.input_dir.folder_path):
            raise NotDirectoryError(
                msg=f"{self.input_dir.folder_path} is not a valid directory.",
                source=self.__class__.__name__,
            )
        check_int(name="TIME WINDOW", value=self.time_window.entry_get, min_value=1)
        _ = Smooth(
            config_path=self.config_path,
            input_path=self.input_dir.folder_path,
            time_window=self.time_window.entry_get,
            smoothing_method=self.method_dropdown.getChoices(),
        )



import argparse
import os.path
import pickle
from copy import deepcopy
from typing import Dict, Optional, Union

import cv2
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.statistics_mixin import Statistics
from simba.utils.read_write import read_df

CONFIG_PATH = "/Users/simon/Desktop/envs/simba/troubleshooting/RI/project_folder/project_config.ini"
VIDEO_NAME = "RI_01_179"


config = ConfigReader(config_path=CONFIG_PATH, create_logger=False)
data_path = os.path.join(config.features_dir, f"{VIDEO_NAME}.{config.file_type}")
df = read_df(file_path=data_path, file_type=config.file_type)
_, _, fps = config.read_video_info(video_name=VIDEO_NAME)

#
# correlation_coefficient = 0.8
#
# # Covariance matrix based on the correlation coefficient
# covariance_matrix = np.array([[1.0, correlation_coefficient],
#                               [correlation_coefficient, 1.0]])
#
#
# data = np.random.multivariate_normal(mean=[0, 0], cov=covariance_matrix, size=len(df))
# value1 = (data[:, 0] > 0).astype(int)
# value2 = (data[:, 1] > 0).astype(int)
#
# df['White_animal_lateral_threat'] = value1
# df['Black_animal_freezing'] = value2
#
# df.to_csv(data_path)

# COMPARE THE ANGLE DISTRIBUTION OF ONE ANIMAL IN ROLLING 500MS TIME WINDOWS USING INDEPENDENT SAMPLES T-TEST.
# USEFUL TO FIND INSTANCES WHEN ANIMAL RAPIDLY CHANGE ITS TAILBASE-CENTROID-NOSE THREE-POINT ANGLE
data = df["Mouse_1_angle"].values.astype(np.float64)
independent_samples_t = Statistics.rolling_independent_sample_t(
    data=data, time_window=0.5, fps=fps
)
time_most_sig_change = (
    np.argmax(np.abs(independent_samples_t)) / fps
)  # E.G, FIND TIMEPOINT in SECONDS WHEN ANGLE CHANGE IS MOST STATISTICALLY SIGNIFICANT
print(time_most_sig_change)

# COMPUTE THE CORRELATION BETWEEN ANIMAL 1 AND ANIMAL 2 CENTROID MOVEMENT IN THE ENTIRE VIDEO USING PEARSONS R
animal_1_data = df["Movement_mouse_1_centroid"].values.astype(np.float32)
animal_2_data = df["Movement_mouse_2_centroid"].values.astype(np.float32)
pearsons_r = Statistics.pearsons_r(sample_1=animal_1_data, sample_2=animal_2_data)
print(pearsons_r)

# ... OR SPEARMANS RANK
spearmans = Statistics.spearman_rank_correlation(
    sample_1=animal_1_data, sample_2=animal_2_data
)
print(spearmans)

# COMPUTE THE PEARSON AND SPEARMAN CORRELATION BETWEEN ANIMAL 1 AND ANIMAL 2 CENTROID MOVEMENT IN IN SLIDING 500MS WINDOWS.
pearsons_r = Statistics.sliding_pearsons_r(
    sample_1=animal_1_data,
    sample_2=animal_2_data,
    time_windows=np.array([0.5]),
    fps=fps,
)
spearman = Statistics.sliding_spearman_rank_correlation(
    sample_1=animal_1_data,
    sample_2=animal_2_data,
    time_windows=np.array([0.5]),
    fps=fps,
)
pearson = np.min(
    np.abs(pearsons_r[:, 0])
)  # GET THE PEARSON R WHEN THE CORRELATION IN 500MS WINDOWS ARE THE LOWEST
data = pd.DataFrame([pearsons_r.flatten(), spearman.flatten()]).T
data.columns = ["pearson", "spearman"]
data.plot(y=["pearson", "spearman"])
# plt.show()

# ... or the two sample KS test comparing the two distributions
results = Statistics.two_sample_ks(sample_1=animal_1_data, sample_2=animal_2_data)

# ... or one way ANOVA
critical_values = pickle.load(
    open(
        "/Users/simon/Desktop/envs/simba/simba/simba/assets/lookups/critical_values_05.pickle",
        "rb",
    )
)["f"]["one_tail"].values.astype(np.float32)
results = Statistics.one_way_anova(
    sample_1=animal_1_data, sample_2=animal_2_data, critical_values=critical_values
)


# Compares the feature values in current time-window to prior time-windows to find the length in time to the most recent time-window where a significantly different feature value distribution is detected.
# See this image as an example.
animal_data = df["Movement_mouse_1_centroid"].values.astype(np.float32)
critical_values = pickle.load(
    open(
        "/Users/simon/Desktop/envs/simba/simba/simba/assets/lookups/critical_values_05.pickle",
        "rb",
    )
)["independent_t_test"]["one_tail"].values.astype(np.float32)
results = Statistics.sliding_independent_samples_t(
    data=animal_data,
    time_window=0.5,
    slide_time=1.0,
    critical_values=critical_values,
    fps=fps,
)

# Compute the sliding auto-correlations (the correlation of the feature with itself using lagged windows).
white_animal = df["Movement_mouse_1_centroid"].values.astype(np.float32)
black_animal = df["Movement_mouse_2_centroid"].values.astype(np.float32)
autocor_white_animal = Statistics.sliding_autocorrelation(
    data=white_animal, max_lag=0.5, time_window=1.0, fps=fps
)
autocor_black_animal = Statistics.sliding_autocorrelation(
    data=black_animal, max_lag=0.5, time_window=1.0, fps=fps
)

# Calculate sliding Z-scores for a given data array over specified time windows.
z_scores = Statistics.sliding_z_scores(
    data=white_animal, time_windows=np.array([0.5, 1.5]), fps=fps
)

# Next we look at the classifications. And compute the chi square correlation coefficent
lateral_threat = df[["White_animal_lateral_threat"]].values.flatten().astype(np.float32)
freezing = df[["Black_animal_freezing"]].values.flatten().astype(np.float32)
chi_square = Statistics.chi_square(sample_1=lateral_threat, sample_2=freezing)[0]

# ... or yule coefficient
yule = Statistics.yule_coef(
    x=lateral_threat.astype(np.int8), y=freezing.astype(np.int8)
)
print(yule)

# ... sokal sneath, using the pose-estimation probability as weights
p_cols = [x for x in df.columns if x.endswith("_p")]
p_vals = np.array(df[p_cols].mean(axis=1))
sokal_sneath = Statistics.sokal_sneath(
    x=lateral_threat.astype(np.int8), y=freezing.astype(np.int8), w=p_vals
)



__author__ = "Simon Nilsson"

import functools
import glob
import itertools
import multiprocessing
import os
import platform
import time
from typing import List, Union

import cv2
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans

from simba.mixins.config_reader import ConfigReader
from simba.utils.data import detect_bouts
from simba.utils.errors import CountError, NoFilesFoundError, NoROIDataError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_core_cnt, find_video_of_file,
                                    get_fn_ext, get_video_meta_data, read_df,
                                    write_df)
from simba.utils.warnings import NoDataFoundWarning


def get_intensity_scores_in_rois(
    frm_list: List[int],
    video_path: Union[str, os.PathLike],
    rectangles_df: pd.DataFrame,
    polygon_df: pd.DataFrame,
    circles_df: pd.DataFrame,
):
    cap = cv2.VideoCapture(video_path)
    start, end = frm_list[0], frm_list[-1]
    cap.set(1, start)
    frm_cnt = start
    results_dict = {}
    while frm_cnt <= end:
        _, img = cap.read()
        results_dict[frm_cnt] = {}
        for idx, rectangle in rectangles_df.iterrows():
            roi_image = img[
                rectangle["topLeftY"] : rectangle["Bottom_right_Y"],
                rectangle["topLeftX"] : rectangle["Bottom_right_X"],
            ]
            results_dict[frm_cnt][rectangle["Name"]] = int(
                np.average(np.linalg.norm(roi_image, axis=2)) / np.sqrt(3)
            )
        for idx, polygon in polygon_df.iterrows():
            x, y, w, h = cv2.boundingRect(polygon["vertices"])
            roi_img = img[y : y + h, x : x + w].copy()
            pts = polygon["vertices"] - polygon["vertices"].min(axis=0)
            mask = np.zeros(roi_img.shape[:2], np.uint8)
            cv2.drawContours(mask, [pts], -1, (255, 255, 255), -1, cv2.LINE_AA)
            dst = cv2.bitwise_and(roi_img, roi_img, mask=mask)
            bg = np.ones_like(roi_img, np.uint8)
            cv2.bitwise_not(bg, bg, mask=mask)
            roi_image = bg + dst
            results_dict[frm_cnt][polygon["Name"]] = int(
                np.average(np.linalg.norm(roi_image, axis=2)) / np.sqrt(3)
            )
        for idx, circle in circles_df.iterrows():
            roi_img = img[
                circle["centerY"] : (circle["centerY"] + 2 * circle["radius"]),
                circle["centerX"] : (circle["centerX"] + 2 * circle["radius"]),
            ]
            mask = np.zeros(roi_img.shape[:2], np.uint8)
            circle_img = cv2.circle(
                mask,
                (circle["centerX"], circle["centerY"]),
                circle["radius"],
                (255, 255, 255),
                thickness=-1,
            )
            dst = cv2.bitwise_and(roi_img, roi_img, mask=circle_img)
            bg = np.ones_like(roi_img, np.uint8)
            cv2.bitwise_not(bg, bg, mask=mask)
            roi_image = bg + dst
            results_dict[frm_cnt][circle["Name"]] = int(
                np.average(np.linalg.norm(roi_image, axis=2)) / np.sqrt(3)
            )
        frm_cnt += 1
    return results_dict


class CueLightAnalyzer(ConfigReader):
    """
    Analyze when cue lights are in ON and OFF states. Results are stored in the
    ``project_folder/csv/cue_lights`` cue lights directory.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str in_dir: directory holding pose-estimation data. E.g., ``project_folder/csv/outlier_corrected_movement_location``
    :parameter List[str] cue_light_names: Names of cue light ROIs, as defined in the SimBA ROI interface.

    .. note::
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.

    Examples
    ----------
    >>> cue_light_analyzer = CueLightAnalyzer(config_path='MyProjectConfig', in_dir='project_folder/csv/outlier_corrected_movement_location', cue_light_names=['Cue_light'])
    >>> cue_light_analyzer.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        in_dir: Union[str, os.PathLike],
        cue_light_names: List[str],
    ):
        ConfigReader.__init__(self, config_path=config_path)

        if len(cue_light_names) == 0:
            raise CountError(msg="SIMBA ERROR: Please select one or more cue lights")
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

        self.out_dir = os.path.join(self.project_path, "csv", "cue_lights")
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)
        self.cue_light_names, self.in_dir = cue_light_names, in_dir
        self.files_found = glob.glob(self.in_dir + "/*" + self.file_type)
        if len(self.files_found) == 0:
            raise NoFilesFoundError(
                msg='SIMBA ERROR: Zero tracking files detected in the "project_folder/csv/outlier_corrected_movement_location" directory'
            )
        _, self.cpu_cnt_to_use = find_core_cnt()
        self.maxtasksperchild = 10
        self.chunksize = 1
        self.read_roi_dfs()
        print(
            "Processing {} cue light(s) in {} data file(s)...".format(
                str(len(cue_light_names)), str(len(self.files_found))
            )
        )

    def read_roi_dfs(self):
        if not os.path.isfile(
            os.path.join(self.logs_path, "measures", "ROI_definitions.h5")
        ):
            raise NoROIDataError(
                msg="No ROI definitions were found in your SimBA project. Please draw some ROIs before analyzing your ROI data"
            )
        else:
            self.roi_h5_path = os.path.join(
                self.logs_path, "measures", "ROI_definitions.h5"
            )
            self.rectangles_df = pd.read_hdf(self.roi_h5_path, key="rectangles")
            self.circles_df = pd.read_hdf(self.roi_h5_path, key="circleDf")
            self.polygon_df = pd.read_hdf(self.roi_h5_path, key="polygons")
            self.shape_names = list(
                itertools.chain(
                    self.rectangles_df["Name"].unique(),
                    self.circles_df["Name"].unique(),
                    self.polygon_df["Name"].unique(),
                )
            )

    def calculate_descriptive_statistics(self):
        self.light_descriptive_statistics = {}
        for shape_name in self.cue_light_names:
            self.light_descriptive_statistics[shape_name] = {}
            self.light_descriptive_statistics[shape_name]["frame_by_frame"] = np.full(
                (self.video_meta_data["frame_count"]), -1
            )
            self.light_descriptive_statistics[shape_name]["descriptive_statistics"] = {}
        for cnt, (k, v) in enumerate(self.intensity_results.items()):
            for shape_name in self.cue_light_names:
                self.light_descriptive_statistics[shape_name]["frame_by_frame"][cnt] = (
                    v[shape_name]
                )
        for shape_name in self.cue_light_names:
            self.light_descriptive_statistics[shape_name]["kmeans"] = KMeans(
                n_clusters=2, random_state=0
            ).fit_predict(
                self.light_descriptive_statistics[shape_name]["frame_by_frame"].reshape(
                    -1, 1
                )
            )
            for i in list(range(0, 2)):
                self.light_descriptive_statistics[shape_name]["descriptive_statistics"][
                    "mean_cluster_{}".format(str(i))
                ] = np.mean(
                    self.light_descriptive_statistics[shape_name]["frame_by_frame"][
                        np.argwhere(
                            self.light_descriptive_statistics[shape_name]["kmeans"] == i
                        ).flatten()
                    ]
                )
                self.light_descriptive_statistics[shape_name]["descriptive_statistics"][
                    "std_cluster_{}".format(str(i))
                ] = np.std(
                    self.light_descriptive_statistics[shape_name]["frame_by_frame"][
                        np.argwhere(
                            self.light_descriptive_statistics[shape_name]["kmeans"] == i
                        ).flatten()
                    ]
                )
            self.light_descriptive_statistics[shape_name]["ON_kmeans_cluster"] = int(
                max(
                    self.light_descriptive_statistics[shape_name][
                        "descriptive_statistics"
                    ],
                    key=self.light_descriptive_statistics[shape_name][
                        "descriptive_statistics"
                    ].get,
                )[-1]
            )
            self.light_descriptive_statistics[shape_name]["ON_FRAMES"] = np.argwhere(
                self.light_descriptive_statistics[shape_name]["kmeans"]
                == self.light_descriptive_statistics[shape_name]["ON_kmeans_cluster"]
            ).flatten()

    def remove_outlier_events(self, frame_threshold=2):
        for cue_light_name in self.cue_light_names:
            que_light_bouts = detect_bouts(
                data_df=self.data_df, target_lst=[cue_light_name], fps=int(self.fps)
            )
            que_light_bouts["frame_length"] = (
                que_light_bouts["End_frame"] - que_light_bouts["Start_frame"]
            )
            que_light_negative_outliers = que_light_bouts[
                que_light_bouts["frame_length"] <= frame_threshold
            ]
            self.que_light_inliers = que_light_bouts[
                que_light_bouts["frame_length"] > frame_threshold
            ]
            for idx, r in que_light_negative_outliers.iterrows():
                self.data_df.loc[
                    r["Start_frame"] - 1 : r["End_frame"] + 1, cue_light_name
                ] = 0

    def insert_light_data(self):
        for shape_name in self.cue_light_names:
            self.data_df.loc[
                list(self.light_descriptive_statistics[shape_name]["ON_FRAMES"]),
                shape_name,
            ] = 1
        self.data_df = self.data_df.fillna(0)

    def run(self):
        start_time = time.time()
        for file_cnt, file_path in enumerate(self.files_found):
            self.data_df = read_df(file_path, self.file_type)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.save_path = os.path.join(
                self.out_dir, self.video_name + "." + self.file_type
            )
            video_settings, pix_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.video_recs = self.rectangles_df.loc[
                (self.rectangles_df["Video"] == self.video_name)
                & (self.rectangles_df["Name"].isin(self.cue_light_names))
            ]
            self.video_circs = self.circles_df.loc[
                (self.circles_df["Video"] == self.video_name)
                & (self.circles_df["Name"].isin(self.cue_light_names))
            ]
            self.video_polys = self.polygon_df.loc[
                (self.polygon_df["Video"] == self.video_name)
                & (self.polygon_df["Name"].isin(self.cue_light_names))
            ]
            self.shape_names = list(
                itertools.chain(
                    self.rectangles_df["Name"].unique(),
                    self.circles_df["Name"].unique(),
                    self.polygon_df["Name"].unique(),
                )
            )
            self.video_path = find_video_of_file(self.video_dir, self.video_name)
            self.video_meta_data = get_video_meta_data(self.video_path)
            if (
                len(self.video_recs) + len(self.video_circs) + len(self.video_polys)
                == 0
            ):
                NoDataFoundWarning(
                    msg=f"No roi data found for video {self.video_name}. Skipping analysis of {self.video_name}..."
                )
                continue

            self.frm_lst = list(range(0, self.video_meta_data["frame_count"], 1))
            self.frame_chunks = np.array_split(
                self.frm_lst, int(self.video_meta_data["frame_count"] / self.fps)
            )
            imgs_peer_loop = len(self.frame_chunks[0])
            self.intensity_results = {}

            with multiprocessing.pool.Pool(
                self.cpu_cnt_to_use, maxtasksperchild=self.maxtasksperchild
            ) as pool:
                functools.partial(
                    get_intensity_scores_in_rois,
                    b=self.video_recs,
                    c=self.circles_df,
                    d=self.video_polys,
                )
                constants = functools.partial(
                    get_intensity_scores_in_rois,
                    video_path=self.video_path,
                    rectangles_df=self.video_recs,
                    circles_df=self.circles_df,
                    polygon_df=self.video_polys,
                )

                for cnt, result in enumerate(
                    pool.imap(constants, self.frame_chunks, chunksize=self.chunksize)
                ):
                    self.intensity_results.update(result)
                    print(
                        "Image {}/{}, Video {}/{}...".format(
                            str(int(imgs_peer_loop * cnt)),
                            str(len(self.data_df)),
                            str(file_cnt + 1),
                            str(len(self.files_found)),
                        )
                    )
            pool.terminate()
            pool.join()
            self.calculate_descriptive_statistics()
            self.insert_light_data()
            self.remove_outlier_events()
            write_df(self.data_df, self.file_type, self.save_path)
        elapsed_time = str(round(time.time() - start_time, 2)) + "s"
        stdout_success(
            msg=f"Analysed {str(len(self.files_found))} files. Data stored in project_folder/csv/cue_lights",
            elapsed_time=elapsed_time,
        )


# test = CueLightAnalyzer(config_path='/Users/simon/Desktop/troubleshooting/light_analyzer/project_folder/project_config.ini',
#                         in_dir='/Users/simon/Desktop/troubleshooting/light_analyzer/project_folder/csv/outlier_corrected_movement_location',
#                         cue_light_names=['Cue_light'])
# test.run()



__author__ = "Simon Nilsson"

import threading
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.clf_validator import ClassifierValidationClips
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.checks import check_int
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.read_write import get_file_name_info_in_directory


class ClassifierValidationPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="SIMBA CLASSIFIER VALIDATION CLIPS")
        ConfigReader.__init__(self, config_path=config_path)
        color_names = list(self.colors_dict.keys())
        self.one_vid_per_bout_var = BooleanVar(value=False)
        self.one_vid_per_video_var = BooleanVar(value=True)
        self.files_found_dict = get_file_name_info_in_directory(
            directory=self.machine_results_dir, file_type=self.file_type
        )
        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.CLF_VALIDATION.value,
        )
        self.seconds_entry = Entry_Box(
            self.settings_frm, "SECONDS PADDING: ", "20", validation="numeric"
        )
        self.clf_dropdown = DropDownMenu(
            self.settings_frm, "CLASSIFIER: ", self.clf_names, "20"
        )
        self.clr_dropdown = DropDownMenu(
            self.settings_frm, "TEXT COLOR: ", color_names, "20"
        )
        self.highlight_clr_dropdown = DropDownMenu(
            self.settings_frm, "HIGHLIGHT TEXT COLOR: ", ["None"] + color_names, "20"
        )
        self.video_speed_dropdown = DropDownMenu(
            self.settings_frm, "VIDEO SPEED: ", Options.SPEED_OPTIONS.value, "20"
        )
        self.clf_dropdown.setChoices(self.clf_names[0])
        self.clr_dropdown.setChoices("Cyan")
        self.seconds_entry.entry_set(val=2)
        self.highlight_clr_dropdown.setChoices("None")
        self.video_speed_dropdown.setChoices(1.0)
        self.individual_bout_clips_cb = Checkbutton(
            self.settings_frm,
            text="CREATE ONE CLIP PER BOUT",
            variable=self.one_vid_per_bout_var,
        )
        self.individual_clip_per_video_cb = Checkbutton(
            self.settings_frm,
            text="CREATE ONE CLIP PER VIDEO",
            variable=self.one_vid_per_video_var,
        )

        self.run_frm = LabelFrame(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_frm = LabelFrame(
            self.run_frm,
            text="SINGLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_btn = Button(
            self.run_single_video_frm,
            text="Create single video",
            fg="blue",
            command=lambda: self.run(multiple_videos=False),
        )
        self.single_video_dropdown = DropDownMenu(
            self.run_single_video_frm,
            "Video:",
            list(self.files_found_dict.keys()),
            "12",
        )
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])
        self.run_multiple_videos = LabelFrame(
            self.run_frm,
            text="MULTIPLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_multiple_video_btn = Button(
            self.run_multiple_videos,
            text="Create multiple videos ({} video(s) found)".format(
                str(len(list(self.files_found_dict.keys())))
            ),
            fg="blue",
            command=lambda: self.run(multiple_videos=True),
        )
        self.settings_frm.grid(row=0, sticky=NW)
        self.seconds_entry.grid(row=0, sticky=NW)
        self.clf_dropdown.grid(row=1, sticky=NW)
        self.clr_dropdown.grid(row=2, sticky=NW)
        self.highlight_clr_dropdown.grid(row=3, sticky=NW)
        self.video_speed_dropdown.grid(row=4, sticky=NW)
        self.individual_bout_clips_cb.grid(row=5, column=0, sticky=NW)
        self.individual_clip_per_video_cb.grid(row=6, column=0, sticky=NW)

        self.run_frm.grid(row=1, column=0, sticky=NW)
        self.run_single_video_frm.grid(row=0, column=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)
        self.run_multiple_videos.grid(row=1, column=0, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, column=0, sticky=NW)

        # self.main_frm.mainloop()

    def run(self, multiple_videos: bool):
        check_int(name="CLIP SECONDS", value=self.seconds_entry.entry_get)
        if self.highlight_clr_dropdown.getChoices() == "None":
            highlight_clr = None
        else:
            highlight_clr = self.colors_dict[self.highlight_clr_dropdown.getChoices()]
        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [
                self.files_found_dict[self.single_video_dropdown.getChoices()]
            ]

        clf_validator = ClassifierValidationClips(
            config_path=self.config_path,
            window=int(self.seconds_entry.entry_get),
            clf_name=self.clf_dropdown.getChoices(),
            clips=self.one_vid_per_bout_var.get(),
            text_clr=self.colors_dict[self.clr_dropdown.getChoices()],
            highlight_clr=highlight_clr,
            video_speed=float(self.video_speed_dropdown.getChoices()),
            concat_video=self.one_vid_per_video_var.get(),
            data_paths=data_paths,
        )
        threading.Thread(target=clf_validator.run()).start()


# _ = ClassifierValidationPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
